(* ElGamal encryption (an RF* example), 
   featuring multiple keys, multiple encryptions, and key-hiding *)

(* SECURITY PROOF

Notations: 

A and IA range over modules and module interfaces
A . B is module composition
I' |- E : I is module typing
"_a" selects a variant of a module definition (#ifdef) 

We write J |- E ~Adv E' : I when, for all A such that I |- A : bool we have 

   | Pr[ E . A = 0 ] - Pr[ E'. A = 0 ] | <= Adv(q,t) 

where t bounds the runtime of A and q bounds the usage of I. 

We ignore the libraries used by all modules, e.g. Bytes, Bignum, Group.

Ienc is an interface guaranteeing abstraction for the private keys (x)
Icpa is an interface guaranteeing also excluding chosen-ciphertext attacks and guaranteeing perfect secrecy on encryptions, 

Definition: AdvDDH(q,t) is such that 

(D)  |- DDH_xy ~AdvDDH DDH_z : IDDH

Theorem: there exists an ideal variant of Elg, Elg_z, such that 

(3)  |- Elg_z                  : Icpa  
(T)  |- Elg ~AdvDDH(q,t) Elg_z : Ienc


The proof involves a series of typechecking judgments (aka Makefile entries)

(0)      |- Elg_0 ~ DDH_xy . Elg_1 : Ienc   (inlining annotations) 
(1) IDDH |- Elg_1                  : Ienc   (abstracting DH secrets; used to apply DDH)
(2)      |- Elg_z ~ DDH_z  . Elg_1 : Ienc   (inlining annotations)
(3)      |- Elg_z                  : Icpa   (relational, with some sample annotation)

The top-level argument shows (T) by applying type safety of module
composition, soundness of the logic, and transitivity:

For any A such that 

(A) Ienc |- A : bool, 

we have 
  |-          Elg_0 . A ~            DDH_xy . Elg_1 . A : eq bool     from 0, A
  |- DDH_xy . Elg_1 . A ~AdvDDH(q,t) DDH_z  . Elg_1 . A               from 1, A, and D applied to Elg_1 . A 
  |- DDH_z  . Elg_1 . A ~                   . Elg_z . A : eq bool     from 2, A

We apply something like "eq_deno" to compute probabilities 

|  Pr [ ____________ = 0 ] - Pr [ ____________ = 0 ] | for each line above. 

TODO: when do we type classic/relational? 
      how to restrict interfaces?

we could in principle do this top-level argument in Coq, relying on a certified rf*-tc.


Note that the runtime complexity (t) against DDH is for the module
Elg_1 . A, not A, so in principle we should also relate the two.

Independently, we show by typing that a module type implements Icpa
provides perfect CPA security: Let G be the CPA game using interface
Ienc: the advantage of the CPA adversary is null.

Icpa lets the adversary provide an *arbitrary* relational formula at
every key generation, setting the refinement for every plaintext
encrypted with that key.


We also need a proof of functional correctness of decryption, relying
on detailed refinement types. *)

module Lib  

(* Basic modulo arithmetic.

   $int for now; we should really try it out with bigint, nicely defined in .NET

extern reference BigInt {language="C#";
                         dll="mscorlib";
                         namespace="System";
                         classname="Numerics"}
extern BigInt type bigint 

 *)

type eq :: _ = fun ('t::*) => (x:'t {| Eq (L x) (R x) |})
open Bytes 

type bigint = int 
val int2bigint: x:int -> y:bigint{x=y}
type nat = i:bigint{i >= 0}
type pos = i:bigint{i > 0}

val bytes2bigint : bytes -> bigint
(* let bigint_of_bytes (b:bytes) : bigint = *)
(*   Array.fold (fun x y -> bigint 256 * x + (bigint (int(y:byte)))) (bigint 0) b *)
val bigint2bytes : bigint -> bytes

type intmod (n:bigint) = i:bigint{0 <= i && i < n}
(* NS: should this be
logic val Modulo: x:bigint -> n:pos -> m:intmod n { exists p. Eq (x - p*n) m }
*)
logic val Modulo: x:bigint -> n:pos -> m:intmod n { exists p. Eq (x + p*n) m }
val invMod: n:bigint -> x:intmod n -> y:intmod n {1 = ((x * y) % n)}

val sample: n:bigint -> intmod n
val sample_eq: n:eq bigint -> eq (intmod n)

(*
val xGCD : x:nat -> n:nat -> a:nat * b:nat{ 1 = (a*x + b*n)  }
let rec xGCD x n =
  if n = bigint 0 then (bigint 1, bigint 0)
  else
    let (q, r) = ((x / n), (x % n)) in
    let (s, t) = xGCD n r in
    (t, s - q * t)

(* Default is to not export lets only to export val *)

let invMod n x = (* (n:bigint) (x:intmod n) : y:intmod n{1 = (Modulo (x * y) n)} =  *)
  let (a,b) = xGCD x n in
  let r = a % n in
  if r < bigint 0 then r+n else r
*)

end 

module Group
open Lib

val p:eq pos
let p = 2*3*5*7*11*13*17 - 1 (* a large prime group *)

type elt = intmod p

val mlt: x:elt -> y:elt -> r:elt {r = ((x*y) % p)}
let mlt x y = (x * y) % p

val div : elt -> elt -> elt
let div x y = mlt x (invMod p y)

logic val Exp : elt -> nat -> elt

val exp : x:elt -> n:nat -> b:elt{|(Eq (L x) (R x) && Eq (L n) (R n)) ==> Eq (L b) (R b)| && (b=Exp x n)}
let rec exp x n =
  if n = 1 
  then (assume (x=Exp x n); x) (* lifing exp into the logic as Exp using this assume *)
  else
    let d = n / 2 in
    let r = n % 2 in
    let a = exp x d in
    let a2 = mlt a a in
    if (r : int) = 0 (* need an annot here because unification for eq is silly *)
    then 
      let _ = assume (a2=Exp x n) in  (* and this one *)
        a2
    else 
      let b = mlt a2 x in  
      let _ = assume (b=Exp x n) in (* and this one *)
        b

val g:eq elt
let g = 17 % p

end

(*
  private function val Exp(e,x) = exp(e,x) // an abstract logical function

  // needed only for the correctness of decryption
  theorem !e,x,y. Exp(Exp(e,x),y) = Exp(e,Prod(x,y))

  // a function to abstract over the g^xy vs g^z in ElGamal encryption
  val exp_ddh: e:group -> y:y -> z:z -> e':elt { !x,y,z. e = Exp(g,x) => e' = < Exp(gx,y) , Exp(g,z) > }
*)

module DDH
open Lib
open Group

(* exponents are integers modulo q *)

val q:eq pos (* modulus for DH exponents *)
let q = 2*3*5*7*11*13 - 1

type exn = intmod q 
private type x = exn  (* protected by type abstraction *)

(* NS: not sure what you intend by GX. 
   CF: it is there to track that it is an "honest" key 
   it would be defined by the existence of an associated secret key.
 *)
(*$ logic val GX: x -> E *)
type gx = eq elt (*$ { GX(e) } *)

val sample_x: unit -> x * gx  (* defined as gx = Exp(g,x) *)
let sample_x () =
  let x = sample_eq q in
  x, exp g x

val exp_x: elt -> x -> elt (* breaking abstraction; used only for decryption *)
(*$ let qy = ref 0 (* private counter *) *) (* NS: not doing monadic stuff yet *)
             
val sample_gy_gxy: gx:elt -> (gy:elt * gxy:elt{ (* ddh ==> *) exists y z. gy = (Exp g y) && gxy = (Exp gx z)})
let sample_gy_gxy gx =
  (*$ qy := !qy + 1; *)
(* 
#if case1 

  let y = sample q in
    (exp gx y, exp gx y)
  
#else *)

  let y = sample q in  
  let z = sample q in  
  (exp g y, exp gx z)

(* #endif *)
    
end

module ElG
open Lib
open Group
open DDH

(*
let qH: int   = failwith "spec"
let g : group = failwith "spec"
let k : int   = failwith "spec"
type kbytes  = bytes // of size k
let zero : kbytes = failwith "spec"
type key        = secret_key * public_key
type message    = kbytes
type cipher     = group * kbytes
*)

(* #if CPA *)
type plain = elt (* later, try adding some formula parameter or getting plain abstraction *)
(* type plain = eq elt *)

(* TODO exclude CCA by refining cipher with an abstract predicate,
   defined as a table lookup in a private log of previously-encrypted plaintexts *)

type cipher = elt * elt
type sk = x
type pk = gx

(*$ a logical specification of El Gamal encryptions:

  function val Enc(gx,y,m) = (Exp(g,y), Prod(Exp(gx,y),m))
  predicate Encrypted(gx,m,c) = ?y. c = Enc(gx,y,m)

  a bit unclear how to express correctness of decryption, concrete & ideal
  !m,c. Encrypted(gx,m,c) => r = m
 *)

val Gen: unit -> sk * pk
let Gen() = sample_x()

val Enc: pk -> plain -> cipher
let Enc pk m =
  let y = sample q in
  (exp g y, mlt (exp pk y) m)

(* #if z *)
(* val sample_z: m:elt -> z:elt {| Eq (Mul (Exp (L g) (L z)) (L m)) (Mul (Exp (R g) (R z)) (R m)) |}  *)
val sample_gz: m:elt -> gz:elt {| Eq (Mul (L gz) (L m)) (Mul (R gz) (R m)) |} 
(* NS: This parses now, but not sure how to prove it. Need a special rule for sample? 
   CF: yes, the one above; patched, but still not typechecking. *)
(*$
let sample_gz m =
  let z = sample_z m in 
  exp g z *)
(* #endif *)

val Enc_1: pk -> plain -> cipher
let Enc_1 pk m =
  let gy, gxy = sample_gy_gxy pk in (* NS: Fails here ... need a gx. CF = pk *)
  (gy, mlt gxy m)

(* this type guarantees secrecy both for the plaintext and for the
   public key (key-hiding-encryption), which is useful for privacy. *)

val Enc_z: pk -> plain -> eq elt * eq elt (*$ eq cipher does not work *)
let Enc_z pk m =
  let y = sample_eq q in
  let gz = sample_gz (m:elt) in
  (exp g y, mlt gz m)

val Dec: sk -> cipher -> plain
let Dec x c =
  let (gy,v) = c in
  let gxy = exp_x gy x in
  div v gxy

(* TODO Dec_z should be by table lookup in the log *)



module CPA
open Lib
open Group
open ElG

(* By typing, we show that our ideal CPA interface implies perfect IND-CPA-1 *)

(* Two adversaries, implicitly stateful *)
type A1 = eq pk -> eq plain * eq plain
type A2 = (eq elt * eq elt) -> eq bool

let game (a1:A1) (a2:A2) b =
  let (sk,pk) = Gen() in
  let (m0,m1) = a1 pk in
  let m = if b then m0 else m1 in
  let c = Enc_z pk m in
  a2 c


module Test
open Lib
open Group
open ElG

(*$ toplevel pair bindings trigger a "not yet implemented" error. *)

let pk = 
  let (sk,pk) = Gen() in 
  pk 

(*$ how to print? *) 


let _ = print_int (pk) 
let _ = print_int (g)
let _ = 
  let a,b = Enc pk g in print_int a && print_int b



