module Group
type elt
val generator : elt

(* logic  *)val order : i:int{i > 0}
type exponent = i:int{0 <= i && i < order}
logic val mul: exponent -> exponent -> exponent
logic val exp: e:elt -> x:exponent -> elt
assume Exp_assumption: forall e x y. exp (exp e x) y = (exp e (mul x y))
assume Mul_commutes: forall e1 e2. mul e1 e2 = mul e2 e1
(* ... more assumptions about mul *)
end

module HashedElG
open Lib
open Group
open RandomOracle

(* // later, try getting plain abstraction *)
type plain = bytes
type cipher = (elt * bytes)
type secret_key = exponent
type public_key = elt

assume Xor_twice: forall b1 b2. XOR b1 (XOR b1 b2) = b2

val keyGen: unit ->  (secret_key * public_key)
let keyGen () : secret_key * public_key = 
  let sk = sample<Eq exponent> () in 
  (sk, exp generator sk)

val enc: 'P::bytes => bytes => E
        -> pk:public_key{Injective 'P}
        -> m:plain 
        -> c:cipher
let enc ('P::bytes => bytes => E) pk m = 
  let y = sample<Eq exponent> () in
  let h = hashH<'P> (exp pk y) in 
  let (gy, v) = (exp generator y, XOR h m) in 
    (* sanity check. *)
    assert (forall (sk:exponent{pk = exp generator sk}). 
              ((exp gy sk = exp pk y) && (XOR h v = m)));
    (gy,v)
      
val dec: secret_key -> cipher -> plain 
let dec sk (gy, v) = 
  let h = hashH<Eq bytes> (exp gy sk) in 
    XOR h v

end

module INDCPA
open HashedElG

(* Two adversaries, implicitly stateful *)
type A1 = public_key -> plain * plain 
type A2 = cipher -> bool

(* Initial games, as in easycrypt *)
let game0 (a1:A1) (a2:A2) =
  let (sk,pk) = keyGen() in
  let (m0,m1) = a1 pk in
  let c = enc pk m0 in
    a2 c
      
let game1 (a1:A1) (a2:A2) =
  let (sk,pk) = keyGen() in
  let (m0,m1) = a1 pk in
  let (gy, _) = enc pk m1 in
  let c = (gy, sample<Eq bytes> ()) in
    a2 c

val constb:bytes

(* Folding the two games into a single program with flag for the cross-cases *)
val game: A1 -> A2 -> bool -> b:bool{L b = R b || Bad}
let game (a1:A1) (a2:A2) b  =
  let (sk,pk) = keyGen() in
  let (m0,m1) = a1 pk in 
    if b then
      let c = enc<fun b1 b2 => (b2 = XOR m b1)> pk m0 in
        a2 c 
    else
      let c = enc<fun b1 b2 => (b2 = XOR m b1)> pk m1 in
        a2 c 


(* Inlining *)
val game: A1 -> A2 -> bool -> b:bool{L b = R b || Bad}
let game (a1:A1) (a2:A2) b =
  let (sk,pk) = keyGen() in
  let (m0,m1) = a1 pk in 
  let m = if b then m0 else m1
  let y = sample<Eq exponent> () in
  let gy = exp generator y in 
  let h = hashH<fun b1 b2 => b1 = XOR b2 m> (exp pk y) in 
  a2 (gy, XOR h m) 

end
