module CCA2

(** Ideal functionality for CCA2-secure public-key encryption with 
    - multiple keys, some of them statically compromised
    - multiple encryptions and decryptions
    - key hiding (commented out for now)

    The main idea is to treat plaintexts as arbitrarily-refined
    fixed-sized bytes, thus capturing both their integrity (using
    classic refinements) and privacy (using non-eq relational
    refinements). 

    One complication with public-key encryption is that 
    the adversary is also able to encrypt arbitrary data,
    so formally the type of plaintexts must be a supertype of eq plain.

    We currently keep the key management implicit, 
    using concrete principals as key identifiers.

    The functionality is parameterized by 
    - Safe, an ideal predicate specifying safe principals.

    We can tweak/simplify for other purposes, e.g. CPA.

 **)

open Bytes
type id = bytes
(* type plain = b:bytes { Length b = 32 }  *) (* Doesn't work with simplenonce *)
type plain = bytes
type cipher = eq bytes

(* Safety of principals is just abstract, for now.
   Also, omitting keys for now *)

type Safe :: id => E
type prin = eq id 

(* for key hiding:
type prin = p:id { Safe p || (| L p = R p |) } 
 *)

val safe : p:prin -> b:bool{b=true <==> Safe p}

(* Concrete primitives, to be linked with some .NET crypto libraries.
   for key hiding, we will need to use them single-sided *)

val enc: eq prin -> eq plain -> cipher
val dec: eq prin -> cipher -> eq (option plain)
val zero: eq plain

(* The type of plaintext for honest encryptors, 
   with different pre-conditions depending on the safety of the public key *)

type eplain :: _ = (fun (b:prin) ('Q::prin => plain => E) => 
    (p:plain{ (Safe b ==> 'Q b p) && (| (not(Safe (L b))) ==> L p = R p|)})) 

(* In addition, dishonest encryptors can use their own copy of the concrete algorithm. *)

(* covering cases of Safe and unsafe principals, and parameterized by relation 'Q *)
type cca2 ('Q::id => plain => E) = 
    {encrypt: b:prin -> eplain b 'Q -> cipher;
     decrypt: b:prin -> cipher -> option (p:plain {'Q b p || (| L p = R p |) })}
      
logic data type logentry ('Q::id => plain => E) =
             | Entry : b:prin -> p:eplain b 'Q -> c:cipher {Safe b} -> logentry 'Q

val mkCCA2 : 'Q::(id => plain => E) -> unit -> cca2 'Q 
let mkCCA2 ('Q::id => plain => E) (u:unit) : cca2 'Q =
  let log : ref (list (logentry 'Q)) = alloc [] in (* alloc is non-monadic ref *)

  let encrypt (b:prin) (p:eplain b 'Q) : cipher = 
    if safe b 
    then let c = enc b zero in
         let e : logentry 'Q = Entry<'Q> b p c in 
         log := (e::!log); (* ugh. parser *)
         c 
    else enc b p in 

  let decrypt (b:prin) (c:cipher) : option (p:plain{'Q b p || (| L p = R p |)}) = 
    let finder (e:logentry 'Q) : option (f:logentry 'Q{b = Entry_proj_1 f}) =
      match e with 
        | Entry b' _ c' -> 
          if (b=b') && (c=c')
          then 
            Some (e : (f:logentry 'Q{b=Entry_proj_1 f}))
          else None in

    match mapfind finder (!log) with
      | Some (Entry _ p _) -> Some p 
      | _ -> 
        match dec b c with 
          | None -> None
          | Some p -> Some (p : (p:plain{'Q b p || (| L p = R p |)})) in (* needed for deep subtyping eq plain <: ... *)

  {encrypt=encrypt;
   decrypt=decrypt}

end 
