module Group 
open Bytes

type elt (* elements of the group *)
logic val idelt: elt 

(* Treating exponents as int is dodgy; they are integers modulo. *)
type tp (* trapdoor (private param) *)
type public_param (* public param: p, g, h *)
type pparam = eq public_param
logic val trap: public_param -> int (* such that g = h^(trap pp) *) 

type text = int (* exponents for values *)
type opng = int (* exponents for openings *)

type Opening :: _ = fun (lx:text) (rx:text) (lo:opng) (ro:opng) => 
    (Eq (lx + lo) (rx + ro))
type opening (x:text) = o:opng {| Opening (L x) (R x) (L o) (R o) |}

type Openings :: list (text * opng) => list (text * opng) => E
type openings = os:list (text * opng){| Openings (L os) (R os) |}
assume Openings_nil: Openings [] []
assume Openings_cons: (forall lx rx lo ro lrest rrest. 
                         ((Opening lx rx lo ro && Openings lrest rrest)
                          <==>  Openings ((lx,lo)::lrest) ((rx,ro)::rrest)))

(* logical spec of fsts, specialized *)
logic val fsts: list (text * opng) -> list text
assume fsts_nil: fsts [] = []
assume fsts_cons: forall a b rest. fsts ((a,b)::rest) = (a::fsts rest)
(*$ todo: use the trapdoor here, but I get some parsing error.  *)

(* single scalar product *) 
logic val ScalarProduct: list int -> list int -> int
assume SP_nil: (ScalarProduct [] [])=1
assume SP_cons: forall x xs y ys p. 
  (* TODO: fix parser to allow (x*y) x+y etc., in formulas *)
  ScalarProduct (x::xs) (y::ys) = Add (Mul x y) (ScalarProduct xs ys)

logic val Commit: public_param -> text -> opng -> elt
val commit: pp:pparam
  -> x:text
  -> (o:opng * c:eq elt { (| Opening (L x) (R x) (L o) (R o) |) && Commit pp x o = c })
(* 
let commit pp x r = pp.g ^ x * pp.h ^ r 
*)

val verify: pp:pparam
  -> x:text
  -> r:opng
  -> c:elt
  -> b:bool{ b=true ==> c = Commit pp x r }
(* 
let verify = c = pp.g ^ x * pp.h ^ r 
*)

(* this is true only computationally; 
   we should instead specify some computational assumption using up to bad. 

   Given p g h, the computational challenge is to provide 
   x0 o0 \neq x1 o1 such that g^x0 * h^o0 = g^x1 * h^o1
 *)
assume Commit_injective: 
  forall pp x0 x1 r0 r1. 
  (Commit pp x0 r0 = Commit pp x1 r1) ==> x0=x1

(* Can prove this post-condition after implementing commit as in the comment above *)
val mult: c0:elt -> c1:elt -> c:elt {forall pp x0 x1 r0 r1.
    (c0=Commit pp x0 r0 && c1=Commit pp x1 r1) ==>
     c=Commit pp (Add x0 x1) (Add r0 r1)}

val exp: c0:elt -> z:int -> c:elt {forall pp x0 r0.
    c0=(Commit pp x0 r0) ==>
    c=Commit pp (Mul x0 z) (Mul r0 z)}

(* needed for vector commitments *) 
logic val Commits: pp:public_param -> list (text * opng) -> list elt
assume Commits_nil: forall pp xrs.  (Commits pp xrs = []) <==> (xrs = [])
assume Commits_cons: forall pp xrs hd tl. ((Commits pp xrs = (hd::tl)) 
                                           <==> (exists x r xrstl. ((Eq xrs ((x,r)::xrstl))
                                                                    && (hd=Commit pp x r))
                                                                    && (tl=Commits pp xrstl)))

type commitments :: _ = fun ('P::list text => E) (pp:public_param) => 
    (cs:list elt{(forall x xs. 'P (x::xs) ==> ('P xs))
                 && (exists xrs. 'P (fsts xrs) && cs=Commits pp xrs)} )
val scalarExp:
     'P::(list text => E)
  -> pp:pparam
  -> commitments 'P pp
  -> ps:list int
  -> c:elt{exists xs r. 'P xs && c=Commit pp (ScalarProduct xs ps) r}
let rec scalarExp ('P::list text => E) pp cs ps = 
  match cs,ps with
    | (c::cn), (p::pn) ->
        let cn = scalarExp<'P> pp cn pn in
          (* 'P ?xs && cn = Commit pp (SP ?xs pn) ?r *)
        let c1 = exp c p in
          assert (exists x r. c=Commit pp x r);
          (* c=Commit pp !x0 !r0 
             ==> c1 = Commit pp (Mul !x0 p) (Mul !r0 p) *)
          assert (exists x r. c1 = Commit pp (Mul x p) (Mul r p));
          assert (exists xs r. cn = Commit pp (ScalarProduct xs pn) r);
	  let res = mult c1 cn in 
            assert (forall x1 xs r1 rn. 
                      ((c1=Commit pp (Mul x1 p) (Mul r1 p))
                       && (cn=Commit pp (ScalarProduct xs pn) rn))
                      ==> (res=Commit pp (Add (Mul x1 p) (ScalarProduct xs pn)) (Add (Mul r1 p) rn)));
            assume (exists x1 xs r1 rn. (*$ z3 struggles to find the witness, although this should be provable. assuming for now *)
                      res=Commit pp (Add (Mul x1 p) (ScalarProduct xs pn)) (Add (Mul r1 p) rn));
            assume False; (*$ still, unclear how to finish the proof. need more axioms about Commit. admitting for now *)
            res 

  | [],[] -> 
      assume False; (*$ need more axioms about commit; admitted for now *) 
      idelt


(** Metering code **)

module Meter
open Group
open Bytes

(* two tail-closed predicates, tracking authentic readings and rates *) 

type Readings::list int => E
type Rates::list int => E
assume forall x xn. Readings (x::xn) ==> Readings xn
assume forall x xn. Rates (x::xn) ==> Rates xn

(* Meter Signing. 

   We could move this to another generic module, 
   taking Signed as a parameter,
   but then we would have to explicitly turn pp and cs into bytes *)

type Signed :: _ = fun (pp:pparam) (cs:list elt) =>
    (exists xrs. Readings (fsts xrs) && cs = Commits pp xrs)

type dsig = bytes 

val sign: pp:pparam -> cs:eq (list elt){Signed pp cs} -> eq dsig

val verify_meter_signature: pp:pparam
  -> cs:eq (list elt)
  -> eq dsig
  -> b:eq bool{ b=true ==> Signed pp cs }
(* Note, single sided rules for match is a bit risky here...
   but we know that they (L xs) and (R xs) have the same size *)


(* METER *) 
val commits: pp:pparam
  -> xs : list int{Readings xs}
  -> (xrs: openings
      * cs: list elt{xs = fsts xrs && |L cs = R cs| && cs=Commits pp xrs })
  (*$ use eq list ... above? but this leads to some error. NS: huh? *)
let rec commits pp xs = match xs with
  | [] -> ([], [])
  | x::xs ->
      let xrs,cs = commits pp xs in
      let r, c = commit pp x in 
        (Cons<text*opng> (x,r) xrs), (Cons<elt> c cs)
          
val meter: pp:pparam
  -> xs:list int{ Readings xs }
  -> (xrs:openings {xs = fsts xrs} *
        cs:eq (list elt) *
        d:eq dsig{ Commits pp xrs = cs })
let meter pp xs =
  let xrs,cs = commits pp xs in
  (xrs, cs, sign pp cs)

(* USER *)
val sums: xrs:openings
  -> ps:eq (list int)(*{Length ps = Length xrs}*) (*$ todo ... add me *)
  -> (x:int { x = ScalarProduct (fsts xrs) ps } * opening x)
let rec sums xrs ps =
  match xrs,ps with
  | (xr::xrs), (p::ps) -> let x,r = xr in
                          let x0,r0 = sums xrs ps in
                          (x0 + x*p), (r0 + r*p)
  | []       , []      -> 1, 0
(*| _        , _       -> failwith "wrong lengths" *) (*$ todo ... add me *)


(* We actually do not need Readings and Rates to make the payment,
   although they matter to authorize the declassification.  *)
val make_payment: pp:pparam
  -> xrs:openings{ Readings (fsts xrs) }
  -> ps:eq (list int){ Rates ps  
                       &&  (| ScalarProduct (fsts (L xrs)) (L ps) = ScalarProduct (fsts (R xrs)) (R ps) |) }
  -> (eq int * eq int)
let make_payment pp xrs ps =
  let x,o = sums xrs ps in
    (x,o)
      
(* VERIFIER *)
val verify_payment: pp:pparam
  -> ps:eq (list int){Rates ps}
  -> cs:eq (list elt)
  -> s:eq dsig
  -> x:text
  -> r:opng
  -> b:bool{ b=true ==> exists xs. Readings xs && x=(ScalarProduct xs ps)}
let verify_payment pp ps cs s x r =
  if verify_meter_signature pp cs s then
    (* assert (exists xrs. Readings (fsts xrs) && Eq (list elt) cs (Commits pp xrs)); *)
    let c = scalarExp<Readings> pp cs ps in
      verify pp x r c
  else
    false
      
