module Nonce
open Bytes 

(* We verify authentication based on fresh Nonces in RRF;
   this is later used in the "private authentication" example. *)

(* Proof outline:

   (1) idealize public-key encryption, replacing plaintexts between honest principals with 0s. 
   (2) abstract over nonces, with an interface
       - gen()
       - coerce b
       - compare n0 n1

       val noHello: n: eq nonce -> unit { forall a b na. Safe a && Safe b && SayHello a b na ==> na <> n } 
       val noAck:   n: eq nonce -> unit { forall a b na. Safe a && Safe b && SayAck   a b na ==> na <> n } 

       internally, maintain a table of coerced values, to check for late collisions. 

   (3) up to bad on nonce collisions: generated nonces are pairwise-distinct, and distinct from coerced values.
       we need this to exclude good receivers from accepting good nonces for the wrong purpose.

   (4) authentication based on fresh nonces, as typed here. 

 *)



(* neq in the prelude:
type Neq :: _ = fun ('a::* ) (x:'a) (y:'a) => (x <> y) (* (not (Eq 'a x y)) so that can provide 'a *)
 *)

type eq 't = x:'t {| Eq (L x) (R x) |}
type nonce = b:bytes { Length b = 16 } (* we need fixed lengths for encryption *)
type plain = bytes
type cipher = eq bytes  (* we treat all ciphertexts as public *) 

type id = bytes      (* principal identifiers *) 
type Safe:: id => E  (* "honest" identifiers: their decryption keys are abstract, 
                         and they do not leak decryptions *)

(* Related principals are consistently safe on both sides,
   but safe principals need not be equal. *)
type Prin :: _ = fun (p0:id) (p1:id) => ((Safe p0 <==> Safe p1) && (Safe p0 || p0 = p1))
type prin = s:id {| Prin (L s) (R s) |}

(* This is a logical definition of SentTo (could be inlined)
 * recording all messages sent to b by Safe senders a 
 *) 

type tag = eq (b:bytes { Length b = 8 }) 
logic val Hello: tag
logic val Ack  : tag
logic val Error: tag
(* let Hello = utf8 "hello:" *)
(* let Ack   = utf8 "ack  :" *)
(* let Error = utf8 "error:" *)


assume neq Hello Ack

(*$ we need them; unclear why. *)
assume (Length Hello) = 8
assume (Length Ack  ) = 8
assume (Length Error) = 8

(* we use constant values when decryption fails *)
val ErrorNonce: eq nonce
val ErrorPrin:  eq prin

(* begin events for messages *)
type SayHello:: id => id => nonce => E
type SayAck::   id => id => nonce => nonce => E 

(* specifying authentication properties associated with encryption payload formats *) 
type Pat :: id => id => bytes => bytes => bytes => bytes => E
type SentTo :: _ (* prin => plain => E *) =
  fun (x:id) (p:plain) =>
    (  (forall (a:id) (na:nonce).             
         ((p = Concat Hello (Concat a na)) ==> (Safe a && SayHello a x na)))
    && (| forall (lb:id) (rb:id) (lna:nonce) (rna:nonce) (lnb:nonce) (rnb:nonce).
         (Prin lb rb
          && Eq (L p) (Concat Ack (Concat lna lnb)) && Eq (R p) (Concat Ack (Concat rna rnb))
          && (lna <> rna)
          && Safe lb
          && SayHello (L x) lb lna && SayHello (R x) rb rna
          && Pat lb rb lna rna lnb rnb)
      ==> (SayAck (L x) lb lna lnb && SayAck (R x) rb rna rnb)|))


(** a generic cca2 pk encryption interface, parameterized by SentTo;
    should be typecheckable from an IND-CCA2 ideal encryption with
    table lookup etc. TODO.
    
    See cca2.fst
 **)

(* omitting keys and sizes for now *) 
(* no implementations for now *)

(* covering two cases, for honest and dishonest encryptions *) 
val log: ref (list (b:prin * p:plain { (Safe b && SentTo b p) || | (L b = R b && L p = R p) |}  * cipher))
val safe: a:prin -> r:eq bool { r = true <==> Safe a }

(* concrete primitives *)
val enc: eq plain -> cipher
val dec: cipher -> eq (option plain)
val zero: eq plain

val encrypt: 
  b:prin -> 
  p:plain { ( Safe b && SentTo b p) || | L b = R b && L p = R p |} ->
  cipher 

val find: 
  (list (b:prin * p:plain { (Safe b && SentTo b p) || | L b = R b && L p = R p|} * cipher)) -> b:prin -> cipher -> 
  option (p:plain { Safe b && SentTo b p })

val decrypt: 
  b:prin -> cipher -> option(p:plain { ( Safe b && SentTo b p) || | L p = R p |})

val alice: p:prin {Safe p}

(* sample usage in private authentication *)
(* using triples instead of concatenations for messages *)

(*TODO still need to implement concrete formats. 
       note that parsing may fail, and that we need to validate b1 to turn in into a prin *)

val parse_hello: p:plain -> option(a:prin * na:nonce   { p = Concat Hello (Concat a na) })
val parse_ack:   p:plain -> option(na:nonce * nb:nonce { p = Concat Ack   (Concat na nb) })
val format: t:tag -> b0:bytes -> b1:nonce -> p:plain   { p = Concat t (Concat b0 b1) } 

(* some safe principal a initiates a session *) 

(* provable from the absence of collisions on na *)
assume SayHello_injective:(forall a b0 b1 na. (SayHello a b0 na && SayHello a b1 na) ==> b0 = b1)


(*$ TODO using a relational typing of sample on (fun x -> if Safe(L a) then tweak x else x)
    where tweak is both injective and such that tweak x <> x. *)
val nonce_hello: 
  a:prin {Safe a} 
  -> b:prin 
  -> na:nonce { SayHello a b na &&  | (Safe(L a) && Safe(L b)) <=> neq nonce (L na) (R na) |}  
(* let nonce_hello a b =  *)
(*   let na = sample() in *)
(*   assume (SayHello a b na);  *)
(*   na *)

val encrypt_hello:
  a:prin { Safe (a:id) } -> 
  b:prin {| Safe (L b) || Eq id (L a) (R a) |} -> 
  cipher
let encrypt_hello a b =
  let na = nonce_hello a b in
  let p = format Hello a na in
  encrypt b p

(* some safe principal b processes a hello message *) 
val decrypt_hello:
  b:prin {Safe b} -> (* friend: prin list -> *)
  cipher -> 
  option(a:prin * na:nonce { (Safe a && SayHello a b na) || | L a = R a && L na = R na |})
  (* at this point b doesn't know much *)
let decrypt_hello b c = 
  match decrypt b c with
    | Some p -> 
      (match parse_hello p with
        | Some((a,na)) -> Some (a,na) 
        | None -> None)
    | None -> None 
      
val nonce_ack: 
  a:prin -> b:prin {Safe b} -> na:nonce ->
  nb:nonce { 
    SayAck a b na nb  && 
  | (Safe(L a) && Safe(L b)) <=> L nb <> R nb |}  

val encrypt_ack: 
  a:prin -> 
  b:prin {Safe b}  ->
  na:nonce { 
      ( Safe a && SayHello a b na              && | neq (L na) (R na)  |) || 
    | ( Eq id (L a) (R a) && Eq id (L b) (R b) &&    Eq (L na) (R na)) |} -> 
  cipher

let encrypt_ack a b na = 
  let nb = nonce_ack a b na in
  let p = format Ack na nb in
  assert (SentTo a p);
  encrypt a p

(* some safe principal a receives a message from some possibly unsafe b *)
(* old bug: {Safe a} or any other refinement may cause a projection error in the postcondition *)
(*$ missing cross-cases for now? we can actually prove they are empty. *)
val decrypt_ack: 
  a:prin {Safe a} -> 
  b:prin ->
  na:nonce  -> 
  c:cipher { SayHello a b na && | ((Safe (L b)) <=> (neq nonce (L na) (R na))) |} ->  (*$ fix me later *)
  option(nb:nonce { Safe b ==> (SayAck a b na nb) })  
let decrypt_ack a b na c = 
  match decrypt a c with
    | Some p -> 
      (match parse_ack p with 
        | Some na_nb -> 
          let (na',nb) = na_nb in 
          if na = na' (* Nonce.compare na na' *)
          then 
            (assume (| Pat (L b) (R b) (L na') (R na') (L nb) (R nb) |); 
             Some nb)
          else None
        | None -> None)
    | None -> None


(*$ to be completed below *) 
(*
val initiator: 
  a:prin { Safe a } -> 
  b:prin -> 
  option(na:nonce * nb:nonce { (| L a = R a|) || (Safe b ==> SayAck a b na nb) })

let initiator a b = 
  (* from encrypt_hello *) 
  let na = nonce_hello a b in
  let p = format Hello a na in
  let e = encrypt b p in
  (* communicate here *)
  (* from decrypt_ack *)
  let p = decrypt a e in
  let (tag,na',nb) = parse_ack p in
  if tag = Ack then 
    if na = na' 
    then Some(na,nb)
    else None
  else None
 *)

(* The responder is parameterized by an arbitrary filter function,
   which could be implemented e.g. as a list of friends. 
   We require that this function be "low-equivalent" on unsafe principals. *)
type filter = a:prin -> r:bool { Safe(a) || (|L r = R r |)}

let decoy = (ErrorPrin, ErrorNonce)
assume Safe ErrorPrin

val responder: b:prin { Safe b } -> filter -> cipher -> (prin * nonce * nonce * cipher)
let responder b f c = 
  let a,na : (a:prin * na:nonce { (Safe a && SayHello a b na) || | L a = R a && L na = R na |})
    = (* each test has single-sided cases, but the two inner function calls are classic  *)
    (match decrypt_hello b c with 
      | Some ((a,na)) -> if f a then (a,na) else decoy
      | None -> decoy ) in
  (* below, we treat "ErrorPrin" as a Safe, silent principal *)
  (* let e0 = encrypt_ack a b na in *)
  let nb = nonce_ack a b na in
  let p = format Ack na nb in
  assert (SentTo a p);
  let e = encrypt a p in
  (a,na,nb,e)
    
