
(* linkability attack against French passports
   http://www.cs.bham.ac.uk/~tpc/Papers/PassportTrace.pdf

   in their notation, using abstract authenticated encryption 
   instead of their encrypt-and-MAC, the basic access control protocol is:


              Passport            sharing K               Reader
                              (optical reading)

                             <--get-challenge---
            NT <-$- 64 bits
                             -------- NT ------>    NR, KR <-$ 64 bits
                             <---{NR NT KR}_K---
            verify decryption
            verify NT
            KT <-$- 64 bits 
                             ----{NT NR KT}_K-->
                                                    verify decryption
                                                    verify NR [not NT?] 

                  now sharing the session key KT xor KR

   several passports (notably the French ones) return different error messages
   whether the first or second check fails, leading to a traceability attack:

   * an adversary can collect encrypted messages associated with real-world identity
   * to check whether a passport belongs to some previously-collected identity, 
     it suffices to try out the old message and observe the error message.
  
 *)

(* simplified, to show the attack and privacy once patched; could be more general.  *) 

module Passport
open Bytes 

val send: eq bytes -> unit  
val recv: unit -> eq bytes


(* module AE: authenticated encryption *)

(* this is plain encrypt-then-MAC except for key unlinkability
   when the payload is fresh (not so easy to implement) *)

(* Specifically when the payload matches one of the two binary patterns
   _|NT|_ or NT|_|_ where NT is a passport nonce, 
   the two authenticated encryptions are equal.
   (linearity? we encrypt at most one such message for all good keys)
   (We must fix the sampling on their PRPs.) *)

(* We model encryption as a lazy PRF and we exclude collisions on NT; 
   this enables us to extend the table of two keys (possibly distinct)
   with the same cipher. 

 *)

type key (* abstract, and not particularly equal *)

type nonce   = eq bytes (* fbytes 16 *)
type payload = eq bytes (* session-key materials *) 
type plain   = bytes (* SAME on both side for this example *)
type cipher  = eq bytes

type Log:: key => key => bytes => bytes => E 

val gen: unit -> key (* no collisions *) 
val encrypt: k:key -> p:plain  -> cipher       (* TODO. adding (k, p, c) to the log. *) 
val decrypt: k:key -> c:cipher -> o:option plain 
     {| forall p. (L o = Some p <==> Log (L k) (R k) (L p) (L c)) && 
                  (R o = Some p <==> Log (L k) (R k) (L p) (L c)) |} 
     (* returning p iff (k,p,c) in the log *)

val errorNonce: cipher
val errorMAC  : cipher

logic val NT: bytes -> nonce (* abstract formatting *)

val nonce: unit -> eq nonce (* ... and (k,p,c) in the log ===> not NT p = nt *)
val data:  unit -> eq payload 

val split: p:plain -> nt:nonce { NT p = nt } * nonce * payload    
val concat: nt:nonce -> nonce -> payload -> p:plain { NT p = nt } 


(* on the secure reader *)

val connect: k:key -> nt:nonce -> cipher 

let connect (k: key) (nt:nonce) = 
  let nr = nonce() in
  let kr = data() in 
  let plain = concat nt nr kr in
  encrypt k plain

(* don't care what happens next:
  let cipher = recv() in
  match AE.DEC k cipher with 
    | Some plain -> 
        let (NT',NR',KT) = split plain in
        if NT = NT' && NR = NR' then *)
      


(* on the smartcard chip *)

(* removed prior nonce generation by the smartcard:

val init: key -> nonce key (* ... not in the log *) 

 *)

val accept: 
  k:key -> nt:nonce {| forall k0 k1 p c. ((NT p = L nt && Log k0 k1 p c) ==> (L k = k0 && R k = k1)) |} -> 
  cipher -> cipher

let accept k nt c = 
  match decrypt k c with
  | Some plain -> 
      ( let (nt',nr,kr) = split plain in
        if (nt' : nonce) = nt then 
          let kt = data() in
          encrypt k (concat nr nt kt)
        else errorNonce )
  | None -> errorMAC 

(*$ let's first simulate the encryption log to experiment with single-sided, 
    by axiomatizing its content for typing accept.
 *) 


(* We have tables (k0 ~ k1, p[nt], c) 
   with at most one k0 ~ k1 for each nt (from its freshness up to collision),
   as recorded in the accept precondition.

   On the left, for instance, 
   decryption succeeds iff exists p | (L k, p, c) in L log 
   and, if the check nt = nt' succeeds, using the the precondition, equivalently
                       iff exists p k0 | (k0, p, c) in L log && NT p = L nt 
 
   and then, from the log invariant, this condition does not depend on the side.
 *)
  
(* for the French attack, 
   we can give the adversary oracles

   val create: unit -> principal
   val key: principal -> principal -> key        (* setting keys, left-right oracle *)
   val init: key -> nt key                       (* getting the first message, e.g. generating a nonce nt *) 
   val connect: key -> nt key -> cipher          (* getting the second message, once *) 
   val accept: key -> nt key -> cipher -> cipher (* getting the third message, once *)

 *)



(*
GAME:

  1. We use table lookups in an SPRP (right?)

  2. For good keys and nonces, we guarantee equality of authenticated
  ciphertexts when they include the nonce, even when they use
  different keys. (This is now typable.) 

  3. We treat collision using some up-to-bad condition.
*)

