(* Copyright (c) Microsoft Corporation.  All rights reserved.  *)

(* this variant does not rely on "plain" abstraction *) 

module PKE
open Bytes 

type eq 't = x:'t {| L x =  R x |}

val plainsize:  int
val pkeysize:   int
val ciphersize: int

private type base_skey = bytes

private type skey = SK : pkey -> skey (* that is recorded in the log *)
logic val PKey: skey -> pkey 
type pkey   = b:eq bytes { Length b = pkeysize && exists sk. b = PKey sk } 

type plain  = b:bytes    { Length b = plainsize } 
type cipher = b:eq bytes { Length b = ciphersize } 
type Msg:: pkey => plain => E 

type Encrypted:: pkey => plain => cipher => E 
(* the semantics of Encrypted should be: in the table *)
(* how to express Encrypted pk p c => Msg pk p ? *)

val base_gen: unit -> sk:skey * pk:pkey { pk = Pkey sk } (* could add context *)
(* kept abstract *) 

(* primitive algorithms, assumed to be functionally correct *)
val base_encrypt: pk:pkey -> p:eq plain -> c:cipher { Encrypted pk p c }
val base_decrypt: sk:skey -> c:cipher -> p:eq plain { forall p'. Encrypted (PKey sk) p' c ==> p = p' } 

(* internal table *) 

type state (* for each key *) = 
  skey * ref (list (c:cipher * p:plain { Encrypted (PKey skey) p c } ))
 
val keys: ref (list (pk:pkey * list (c:cipher * p:plain { Encrypted pk p c)))

let log = ref [] (* state for all keypairs generated so far *)
let zero = zeroCreate plainsize

let rec assoc (pk:pkey) ks : state = 
  match ks with 
  | (y,s)::ks -> if pk = y then s else assoc pk ks
  | [] -> failwith "never"

let rec assocc (pk:pkey) (c:cipher) (cxs:((cipher * plain) list)) = 
  match cxs with 
  |  (c',x)::cxs -> if c = c' then Some x else assocc pk c cxs
  |  [] -> None

let empty_log (pk:pkey) : (cipher * plain) list ref = ref []

let gen () = 
  let pk, sk = base_gen () in
  keys := (pk,(sk, empty_log pk))::!keys;  
  (pk,SK(pk))

(* idealized encrypt/decrypt: now plains need not be eq *)
val encrypt: pk:pkey -> p:plain { Msg pk p } -> c:cipher { Encrypted pk p c }
val decrypt: sk:skey -> c:cipher -> p:plain { forall p'. Encrypted (PKey sk) p' c ==> p = p' } 

(* non-malleability means that decryption should guarantee eq on non-Msg *)

let encrypt pk x = 
  let (sk,log) = assoc pk !keys in
  let c = base_encrypt pk zero in 
  log := (cons (c,x) !log); 
  c 

let decrypt (SK(pk)) c = 
  let (sk,log) = assoc pk !keys in
  match assocc pk c !log with
  | Some(x) -> x 
  | None -> base_decrypt sk c            

