module Prims

(* logical connectives *)
type l_and :: E => E => P
type l_or  :: E => E => P
type l_not :: E => P
type l_iff :: E => E => P
type l_implies :: E => E => P
type Forall :: 'a::* => ('a => E) => E
type Exists :: 'a::* => ('a => E) => E
type ForallA :: 'a::A => ('a => E) => E
type ExistsA :: 'a::A => ('a => E) => E
type Relational :: E => E
type DoubleSided :: E => E
type SPEC_ONLY :: E => E
type True :: P
type False :: P
type EqTyp :: E => E => E
type Eq :: 'a::* => 'a => 'a => P
type Eq2 :: 'a::* => 'b::* => 'a => 'b => P
type EqA :: 'a::A => 'a => 'a => E
type TypeOf :: 'a::* => 'a => E
logic tfun type AsE :: 'a::* => 'a => E
type neq :: _ = (fun ('a::*) (x:'a) (y:'a) => l_not (Eq 'a x y))

type NTuple =
  | Tuple_UU : 'a -> 'b -> ('a * 'b)
  | Tuple_UA : 'a::* -> 'b::A -> 'a -> 'b -> ('a * 'b) 
  | Tuple_AU : 'a::A -> 'b::* -> 'a -> 'b -> ('a * 'b)
  | Tuple_AA : 'a::A -> 'b::A -> 'a -> 'b -> ('a * 'b)

logic val fst : ('a * 'b) -> 'a
logic val snd : ('a * 'b) -> 'b

type pf  :: E => P  =
  | T                : pf True

assume fst_def: forall x y. fst (x,y) = x
assume snd_def: forall x y. snd (x,y) = y

type object
type bool
type unit
assume Unit_id: forall (x:unit). x=()
type int
type string
type LBL :: string => E => E
(* type bytes *)
logic data type exn =
  | Exception : string -> exn
type float
type result :: * => * =
  | V : 'a -> result 'a
  | E : exn -> result 'a
  | Err : result 'a
type undef 
type nul 
logic val L : 'a -> 'a
logic val R : 'a -> 'a
(* Primitive (structural) equality. 
   What about for function types? *)
val op_Equality : x:'a -> y:'a -> z:bool {(z=true <==> x=y) && (z=false <==> (x<>y))}
type eq :: _ = fun ('t::*) => (x:'t {| Eq (L x) (R x) |})

(* Integer arithmetic *)
logic val Add : int -> int -> int
logic val Sub : int -> int -> int
logic val Mul : int -> int -> int
logic val Div : int -> int -> int
logic val Minus : int -> int
logic val Modulo : int -> int -> int

type LT :: int => int => E
type GT :: int => int => E
type LTE :: int => int => E
type GTE :: int => int => E

val id : 'a::* -> 'a -> 'a
let id x = x

val idprop : 'a::P -> 'a -> 'a
let idprop x = x

val apply: ('a -> 'b) -> 'a -> 'b
let apply f x = f x

val idint: int -> int
let idint x = id x

logic data type option 'a =
  | None : option 'a
  | Some : 'a -> option 'a

type either 'a 'b =
  | Inl : 'a -> either 'a 'b
  | Inr : 'b -> either 'a 'b

val bind_opt: ('a -> 'b) -> option 'a -> option 'b
let bind_opt f x = match x with
  | None -> None
  | Some x -> Some (f x)
  
logic data type list 'a =
  | Nil : list 'a
  | Cons : 'a -> list 'a -> list 'a

type In :: 'a::* => 'a => list 'a => E
type ListUnion :: 'a::* => list 'a => list 'a => list 'a => E
assume In_hd: forall (hd:'a) (tl:list 'a). (In hd (Cons hd tl))
assume In_tl: forall (hd:'a) (x:'a) (tl:list 'a). (In x tl) => (In x (Cons hd tl))
assume notinNil: forall (x:'a). not (In x Nil)
assume notinCons: forall (x:'a) (y:'a) (tl:list 'a). ((not (In x tl)) && (not (x=y))) => not (In x (Cons y tl))
type ListInclusion :: 'a::* => list 'a => list 'a => E
assume forall l1 l2. ListInclusion l1 l2 <==> (forall x. In x l1 ==> In x l2)

val mem: x:'a -> l:list 'a -> b:bool{b=true <==> In x l}
let rec mem x = function 
  | [] -> false
  | hd::tl -> if hd = x then true else mem x tl

val map: ('a -> 'b) -> list 'a -> list 'b
let rec map f x = match x with 
  | Nil -> Nil
  | Cons a tl -> Cons (f a) (map f tl)

val fold_left: ('a -> 'b -> 'a) -> 'a -> list 'b -> 'a
let rec fold_left f x y = match y with 
  | Nil -> x
  | Cons hd tl -> fold_left f (f x hd) tl

val fold_right: ('a -> 'b -> 'b) -> list 'a -> 'b -> 'b
let rec fold_right f l x = match l with
  | Nil -> x
  | Cons hd tl -> fold_right f tl (f hd x)

val iterate: ('a -> unit) -> list 'a -> unit
let rec iterate f x = match x with
  | Nil -> ()
  | Cons a tl -> let _ = f a in iterate f tl
                                  
val fold_left_A: 'a::A -> 'b::* -> ('a -> 'b >> 'a) -> 'a -> list 'b >> 'a
let rec fold_left_A f a l = match l with
  | Nil -> a
  | Cons hd tl -> fold_left_A f (f a hd) tl
 
val assoc: 'a -> list ('a*'b) -> option 'b
let rec assoc a x = match x with
  | Nil -> None
  | Cons (a', b) tl -> if a=a' then Some b else assoc a tl

val find: ('a -> bool) -> list 'a -> option 'a
let rec find f = function 
  | [] -> None
  | hd::tl -> if f hd then Some hd else find f tl

val mapfind : ('a -> option 'b) -> list 'a -> option 'b
let rec mapfind f = function 
  | [] -> None
  | hd::tl -> match f hd with Some hd' -> Some hd' | _ -> mapfind f tl

(* val append: x:list 'a -> y:list 'a -> z:list 'a { forall (a:'a). In a z <=> (In a x || In a y)} *)
(* let rec append x y = match x with *)
(*   | Nil -> y *)
(*   | Cons a tl -> Cons a (append tl y) *)

(* val concatMap: ('a -> list 'b) -> list 'a -> list 'b *)
(* let rec concatMap f = function *)
(*   | [] -> [] *)
(*   | a::tl -> append (f a) (concatMap f tl) *)

extern reference String {language="C#";
                         dll="mscorlib";
                         namespace="System";
                         classname="String"}

extern String val Concat: string -> string -> string


(* Some library functions *)
extern reference SysConvert {language="C#";
                             dll="mscorlib";
                             namespace="System";
                             classname="Convert"}
(* extern SysConvert val ToBase64String : bytes -> string *)
(* extern SysConvert val FromBase64String : string -> bytes *)

extern reference Runtime { language = "F#";
                           dll="runtime";
                           namespace="Microsoft.FStar.Runtime";
                           classname="Pickler"}

(* type Serialized :: 'a::* => 'a => bytes => E *)

logic val ReprInt: int -> string
logic val Strcat : string -> string -> string
extern Runtime type punit :: P
extern Runtime val freshname : bool -> string
extern Runtime val debug_println : string -> bool
extern Runtime val println : string -> bool
extern Runtime val printfile: string -> string -> bool
extern Runtime val printfileNoLn: string -> string -> bool
extern Runtime val print_stderr : string -> bool
extern Runtime val print_string : string -> bool
extern Runtime val string_of_any : 'a -> string
extern Runtime val string_of_any_for_coq : 'a -> string
extern Runtime val string_of_any_for_coq_afn : 'a -> string
extern Runtime val writeToFile : string -> 'a -> string
extern Runtime val writeCertToFile : string -> 'a -> string
extern Runtime val print_int : int -> bool
extern Runtime val strcat : string -> string -> string
extern Runtime val strStartsWith: string -> string -> bool
extern Runtime val intToString: n:int -> s:string{s=(ReprInt n)}
extern Runtime val stringToInt: s:string -> n:int{s=(ReprInt n)}
extern Runtime val strRmPfx: s:string -> pfx:string -> r:string{s=(Strcat pfx r)}
extern Runtime val intCheckRange: int -> int -> int -> bool

extern Runtime val strSplitByDelimiter: s:string -> d:string -> (r1:string*r2:string{(Strcat r1 r2)=s})
(* extern Runtime val createComm: int -> ((bool -> bytes) * (bytes -> bool)) *)
extern Runtime val stopAllServers: bool -> bool
extern Runtime val keyGen: bool -> (string * string)

extern Runtime val boxToObject: 'a -> object
extern Runtime val addBindings: object -> string -> bool
extern Runtime val lookupBindings: object -> option string
extern Runtime val clearBindings: bool -> bool

extern Runtime val Assume: 'P::E -> unit -> (y:unit{'P})
extern Runtime val Assert : 'P::E -> x:unit{'P} -> z:unit{'P}
extern Runtime val lemma : 'P::E -> x:unit{'P} -> z:unit{'P}
extern Runtime val unreachable : x:unit{False} -> 'a
(* extern Runtime val pickle: x:'a -> (b:bytes{Serialized x b}) *)
(* extern Runtime val unpickle: b:bytes -> (x:'a{Serialized x b}) *)
extern Runtime val throw: string -> 'a 

(* -------------------------------------------------------------------------------- *)
(* The Dijkstra state monad *)
(* -------------------------------------------------------------------------------- *)
(* The primitive heap of references is modeled using a Select/Update theory *)
type ref :: * => *
logic array(SelHeap, UpdHeap, EmpHeap, InHeap) type heap (* = list (loc * obj) *)
logic val SelHeap : 'a::* -> heap -> ref 'a -> 'a
logic val UpdHeap : 'a::* -> heap -> ref 'a -> 'a -> heap    
logic val EmpHeap : heap
type InHeap :: 'a::* => heap => ref 'a => E



(* (RDST 'a 'Tx) is a type of a computation that when given input heap h 
   produces an 'a result x and an output heap h'
   such that, for any relation ('Post :: 'a => 'a => heap => heap => E)
   the weakest pre-condition of the computation to establish ('Post (L x) (R x) (L h') (R h'))
   is ('Tx 'Post (L h) (R h)). 

   That is, 'Tx is a predicate transformer, 
   from relational post-conditions to relational pre-conditions.

   Notice that the definition of RST uses a relational refinement x:t{| phi |}
   on both its argument and result.
*)
type RST :: _ = fun ('Pre::heap => heap => E) ('a::*) ('Post::'a => 'a => heap => heap => E) => 
    h:heap{|'Pre (L h) (R h)|} -> (x:'a * h':heap{|'Post (L x) (R x) (L h') (R h')|})

type RDST :: _ = fun ('a::*) ('Tx::('a => 'a => heap => heap => E) => heap => heap => E) =>
    'Post::('a => 'a => heap => heap => E)
    -> RST ('Tx 'Post) 'a 'Post

type HeapInv :: heap => heap => E
type Bad :: heap => heap => E
(* type DeltaHeap :: heap => heap => heap => heap => E *)
(* assume DeltaHeap_refl: forall h0 h1. DeltaHeap h0 h1 h0 h1 *)
(* assume DeltaHeap_trans: (forall h0 h1 h0' h1' h0'' h1''.  *)
(*                            (DeltaHeap h0 h1 h0' h1' && DeltaHeap h0' h1' h0'' h1'') *)
(*                          ==> DeltaHeap h0 h1 h0'' h1'') *)
type WithInv :: _ = fun 
  ('a::*) 
  ('Tx::('a => 'a => heap => heap => E) => heap => heap => E) 
  ('Post::'a => 'a => heap => heap => E) (h0:heap) (h1:heap) => 
    (HeapInv h0 h1 
     && (('Tx ((fun (r0:'a) (r1:'a) (h0':heap) (h1':heap) => 
                 ((HeapInv h0' h1' (* && DeltaHeap h0 h1 h0' h1' *)) ==> 'Post r0 r1 h0' h1'))) h0 h1)))
type iRDST :: _ = fun ('a::*) ('Tx::('a => 'a => heap => heap => E) => heap => heap => E) =>
    RDST 'a (WithInv 'a 'Tx)
    
type M 'a = iRDST 'a ((fun ('Post::'a => 'a => heap => heap => E) h0 h1 => 
                          (forall r0 r1 h0' h1'. 'Post r0 r1 h0' h1')))

type composeClassic :: _ = (fun ('a::*) 
                              ('Tx1::('a => heap => E) => heap => E) 
                              ('Tx2::('a => heap => E) => heap => E) 
                              ('Post::'a => 'a => heap => heap => E) 
                              h0 h1 => 
    ('Tx1 (fun v0 h0' => 'Tx2 (fun v1 h1' => 'Post v0 v1 h0' h1') h1) h0))
type classicRet :: _ = fun ('a::*) (x:'a) ('Post::'a => heap => E) => 'Post x
type classicBind :: _ = (fun ('a::*) ('b::*) 
                           ('Tx1::('a => heap => E) => heap => E)
                           ('Tx2::'a => ('b => heap => E) => heap => E)
                           ('Post::'b => heap => E) => 
    ('Tx1 (fun (x:'a) => 'Tx2 x 'Post)))

type iRetTx :: _ = fun ('a::*) (x:'a) => (composeClassic 'a (classicRet (L x)) (classicRet (R x)))

type bindTx :: _ = (fun ('a::*) ('b::*) 
                       ('Tx1::('a => 'a => heap => heap => E) => heap => heap => E)
                       ('Tx2::'a => ('b => 'b => heap => heap => E) => heap => heap => E)
                       ('Post::'b => 'b => heap => heap => E)
                       (h0:heap) (h1:heap) => 
    ('Tx1 (fun (y0:'a) (y1:'a) (h0':heap) (h1':heap) => 
        (forall (x:'a). ((L x)=y0 && (R x)=y1) ==> 
           'Tx2 x 'Post h0' h1')) h0 h1))
    
type iBindTx :: _ = (fun ('a::*) ('b::*) 
                       ('Tx1::('a => 'a => heap => heap => E) => heap => heap => E)
                       ('Tx2::'a => ('b => 'b => heap => heap => E) => heap => heap => E)
                       ('Post::'b => 'b => heap => heap => E)
                       (h0:heap) (h1:heap) => 
    (HeapInv h0 h1 
     && ('Tx1 ((fun (y0:'a) (y1:'a) (h0':heap) (h1':heap) => 
                   (HeapInv h0' h1' 
                    (* && DeltaHeap h0 h1 h0' h1' *)
                    && (forall (x:'a). ((L x)=y0 && (R x)=y1) ==> 
                          'Tx2 x ((fun (z0:'b) (z1:'b) (h0'':heap) (h1'':heap) => 
                                      (HeapInv h0'' h1''
                                       (* && DeltaHeap h0' h1' h0'' h1'' *)
                                       && 'Post z0 z1 h0'' h1''))) h0' h1'))))
           h0 h1)))

type DD :: _ = fun ('a::*) ('Tx::('a => heap => E) => heap => E) => 
    RDST 'a (composeClassic 'a 'Tx 'Tx)

type allocTx :: _ = fun ('a::*) (v:'a) ('Post::ref 'a => heap => E) h => 
    (forall (x:ref 'a). not (InHeap h x)
     ==> 'Post x (UpdHeap h x v))
val ref':
     'a::*
  -> v:'a
  -> RDST (ref 'a) (composeClassic (ref 'a) (allocTx 'a (L v)) (allocTx 'a (R v)))
val ref:
     'a::*
  -> v:'a
  -> RDST (ref 'a) (fun ('Post::ref 'a => ref 'a => heap => heap => E) h0 h1 => 
      (forall (x0:ref 'a) (x1:ref 'a). (not (InHeap h0 x0) && not (InHeap h1 x1))
       ==> 'Post x0 x1 (UpdHeap h0 x0 (L v)) (UpdHeap h1 x1 (R v))))

type readTx :: _ = fun ('a::*) (r:ref 'a) ('Post::'a => heap => E) h => 
    ('Post (SelHeap h r) h)
val read': 'a::*
  -> r:ref 'a
  -> RDST 'a (composeClassic 'a (readTx 'a (L r)) (readTx 'a (R r)))
val read: 'a::*
  -> r:ref 'a
  -> RDST 'a (fun ('Post::'a => 'a => heap => heap => E) h0 h1 => 
      'Post (SelHeap h0 (L r)) (SelHeap h1 (R r)) h0 h1)
  
type writeTx :: _ = fun ('a::*) (r:ref 'a) (v:'a) ('Post::unit => heap => E) (h:heap) => 
    ('Post () (UpdHeap h r v))
val write': 'a::*
  -> r:ref 'a
  -> v:'a 
  -> RDST unit (composeClassic unit (writeTx 'a (L r) (L v)) (writeTx 'a (R r) (R v)))
val write: 'a::*
  -> r:ref 'a
  -> v:'a 
  -> RDST unit (fun ('Post::unit => unit => heap => heap => E) h0 h1 => 
      'Post () () (UpdHeap h0 (L r) (L v)) (UpdHeap h1 (R r) (R v)))


type Transitive :: _ = fun ('a::*) ('P::'a => 'a => E) => 
    (forall (x:'a) (y:'a) (z:'a). ('P x y && 'P y z) ==> 'P x z)

(* private *) logic data type mref ('a::*) ('P::'a => 'a => E) :: * = 
  | MRef : r:ref 'a -> mref 'a 'P

logic val SelM : 'a::* -> 'P::'a => 'a => E -> heap -> mref 'a 'P -> 'a
define SelM_def: forall ('a::*) ('P::'a => 'a => E) (h:heap) (m:mref 'a 'P). 
        (Eq (SelM<'a,'P> h m) (SelHeap h (MRef_proj_2 m)))
logic val UpdM : 'a::* -> 'P::'a => 'a => E -> heap -> mref 'a 'P -> 'a -> heap
define UpdM_def: forall ('a::*) ('P::'a => 'a => E) (h:heap) (m:mref 'a 'P) (x:'a). 
        (Eq (UpdM<'a,'P> h m x) (UpdHeap h (MRef_proj_2 m) x))

val newmref: 'a::* 
         -> 'P::('a => 'a => E) 
         -> v:'a{Transitive 'a 'P}
         -> RDST (mref 'a 'P) 
                 (fun ('Post::(mref 'a 'P) => (mref 'a 'P) => heap => heap => E) h0 h1 => 
                     (forall (x0:ref 'a) (x1:ref 'a). 
                        (not (InHeap h0 x0) && not (InHeap h1 x1))
                        ==> 'Post (MRef<'a, 'P> x0) (MRef<'a, 'P> x1) (UpdHeap h0 x0 (L v)) (UpdHeap h1 x1 (R v))))

val mread: 'a::* 
  -> 'P::'a => 'a => E
  -> r:mref 'a 'P
  -> RDST 'a (fun ('Post::'a => 'a => heap => heap => E) h0 h1 => 
      'Post (SelHeap h0 (L (MRef_proj_2 r))) (SelHeap h1 (R (MRef_proj_2 r))) h0 h1)

val mwrite: 'a::*
  -> 'P::'a => 'a => E
  -> r:mref 'a 'P
  -> v:'a 
  -> RDST unit (fun ('Post::unit => unit => heap => heap => E) h0 h1 => 
      ('P (SelM h0 (L r)) (L v)
       && 'P (SelM h1 (R r)) (R v)
       && 'Post () () (UpdM h0 (L r) (L v)) (UpdM h1 (R r) (R v))))

val assert_lemma: 'T::E 
  -> unit
  -> RDST unit (fun ('Post::unit => unit => heap => heap => E) h0 h1 => 
      ('T && ('T ==> 'Post () () h0 h1)))

type Modifies :: _ = fun ('a::*) (x:ref 'a) (h:heap) (h':heap) =>
    (forall ('b::*) (y:ref 'b). (not (Eq2 x y)) ==> (SelHeap h y = SelHeap h' y))
type ModifiesM :: _ = fun ('a::*) ('P::'a => 'a => E) (x:mref 'a 'P) (h:heap) (h':heap) =>
    (forall ('b::*) (xx:ref 'a) (y:ref 'b). ((Eq (MRef_proj_2 x) xx) && (not (Eq2 xx y))) ==> (SelHeap h y = SelHeap h' y))
type Monotonic :: _ = fun (h:heap) (h':heap) => 
    (forall (x:'a) ('P::'a => 'a => E) (m:mref 'a 'P). 
       (InHeap h (MRef_proj_2 m) 
        ==> (InHeap h' (MRef_proj_2 m) && 'P (SelM h m) (SelM h' m))))
(* assume DeltaHeap_inv: (forall h0 h1 h0' h1'. DeltaHeap h0 h1 h0' h1' *)
(*                        ==> (Monotonic h0 h0' && Monotonic h1 h1')) *)

type MonoBoolT :: _ = fun b0 b1 => (b0=true ==> b1=true)
type badmref = mref bool MonoBoolT

type EnsuresUpTo :: _ = fun (b:badmref) ('a::*) ('Ensures::heap => heap => 'a => 'a => heap => heap => E) => 
    RDST 'a (fun ('Post::'a => 'a => heap => heap => E) h0 h1 => 
        (forall a0 a1 h0' h1'. 
           (not ('Ensures h0 h1 a0 a1 h0' h1') ==> (true = SelM h0' (L b) && true = SelM h1' (R b)))
         ==> 'Post a0 a0 h0' h1'))

type iReqEns :: _ = fun ('a::*) ('Req::heap => heap => E) ('Ens::heap => heap => 'a => 'a => heap => heap => E) => 
    (iRDST 'a (fun ('Post::'a => 'a => heap => heap => E) (h0:heap) (h1:heap) => 
        ('Req h0 h1 
         && (forall a0 a1 h0' h1'. 'Ens h0 h1 a0 a1 h0' h1' ==> 'Post a0 a1 h0' h1'))))

type ReqEns :: _ = fun ('a::*) ('Req::heap => heap => E) ('Ens::heap => heap => 'a => 'a => heap => heap => E) => 
    (RDST 'a (fun ('Post::'a => 'a => heap => heap => E) (h0:heap) (h1:heap) => 
        ('Req h0 h1 
         && (forall a0 a1 h0' h1'. 'Ens h0 h1 a0 a1 h0' h1' ==> 'Post a0 a1 h0' h1'))))

type Ensures :: _ = fun ('a::*) ('E::heap => heap => 'a => 'a => heap => heap => E) => 
    (ReqEns 'a (fun h0 h1 => True) 'E)

type iEnsures :: _ = fun ('a::*) ('E::heap => heap => 'a => 'a => heap => heap => E) => 
    (iReqEns 'a (fun h0 h1 => True) 'E)

type iR :: _ = fun ('a::*) ('Req::heap => heap => E) ('Ens::heap => heap => 'a => 'a => heap => heap => E) => 
    (RDST 'a (WithInv 'a ((fun ('Post::'a => 'a => heap => heap => E) (h0:heap) (h1:heap) => 
                               ('Req h0 h1 
                                && (forall a0 a1 h0' h1'. 'Ens h0 h1 a0 a1 h0' h1' ==> 'Post a0 a1 h0' h1'))))))

val get : unit -> iRDST heap (fun ('Post::heap => heap => heap => heap => E) h0 h1 => ('Post h0 h1 h0 h1))
assume HeapTyping: forall h0 h1 (x0:ref 'a) (x1:ref 'a). 
    (HeapInv h0 h1 
     && InHeap h0 x0 
     && InHeap h1 x1) 
    ==> (EqTyp (TypeOf (SelHeap h0 x0)) 'a
         && EqTyp (TypeOf (SelHeap h1 x1)) 'a)

(* Primitive functions with trusted specs  *)
val alloc: 'a -> ref 'a
val _dummy_op_ColonEquals: ref 'a -> 'a -> unit
val _dummy_op_Dereference: ref 'a -> 'a
val _dummy_op_AmpAmp             : x:bool -> y:bool -> z:bool { z=true =>  x=true &&  y=true}
val _dummy_op_BarBar             : x:bool -> y:bool -> z:bool { (z=true => x=true ||  y=true) && 
                                                                 (z=false => x=false && y=false)}
val _dummy_op_Negation           : x:bool -> y:bool { (y=true => x=false) && (y=false => x=true)}

val _dummy_op_Multiply           : x:int -> y:int -> (z:int{z=(x * y)})
val _dummy_op_Division           : x:int -> y:int{y<>0} -> (z:int{z=(x / y)})
val _dummy_op_Subtraction        : x:int -> y:int -> (z:int{z=(x - y)})
val _dummy_op_Addition           : x:int -> y:int -> (z:int{z=(x + y)})
val _dummy_op_Minus              : x:int -> y:int{y=(Minus x)}
val _dummy_op_Modulus            : x:int -> y:int -> z:int{z=(Modulo x y)}

val _dummy_op_GreaterThanOrEqual : x:int -> y:int -> z:bool{((z=true) ==> (x >= y)) && ((z=false) ==> (x < y))}
val _dummy_op_LessThanOrEqual : x:int -> y:int -> z:bool{((z=true) ==> (x <= y)) && ((z=false) ==> (x > y))}
val _dummy_op_GreaterThan : x:int -> y:int -> z:bool{((z=true) ==> (x > y)) && ((z=false) ==> (x <= y))}
val _dummy_op_LessThan : x:int -> y:int -> z:bool{((z=true) ==> (x < y)) && ((z=false) ==> (x >= y))}


