(* This module defines a relational monad for use with inference *)
module RST

(* (RDST 'a 'Tx) is a type of a computation that when given input heap h 
   produces an 'a result x and an output heap h'
   such that, for any relation ('Post :: 'a => 'a => heap => heap => E)
   the weakest pre-condition of the computation to establish ('Post (L x) (R x) (L h') (R h'))
   is ('Tx 'Post (L h) (R h)). 

   That is, 'Tx is a predicate transformer, 
   from relational post-conditions to relational pre-conditions.

   Notice that the definition of RST uses a relational refinement x:t{| phi |}
   on both its argument and result.
*)
type RST :: _ = fun ('Pre::heap => heap => E) ('a::*) ('Post::'a => 'a => heap => heap => E) => 
    -> h:heap{|'Pre (L h) (R h)|} -> (x:'a * h':heap{|'Post (L x) (R x) (L h') (R h')|})

type RDST :: _ = fun ('a::*) ('Tx::('a => 'a => heap => heap => E) => heap => heap => E) =>
    'Post::('a => 'a => heap => heap => E)
    -> RST ('Tx 'Post) 'a 'Post

type HeapInv :: heap => heap => E
type Bad :: heap => heap => E
type DeltaHeap :: heap => heap => heap => heap => E
assume DeltaHeap_refl: forall h0 h1. DeltaHeap h0 h1 h0 h1
assume DeltaHeap_trans: (forall h0 h1 h0' h1' h0'' h1''. 
                           (DeltaHeap h0 h1 h0' h1' && DeltaHeap h0' h1' h0'' h1'')
                         ==> DeltaHeap h0 h1 h0'' h1'')
type WithInv :: _ = fun 
  ('a::*) 
  ('Tx::('a => 'a => heap => heap => E) => heap => heap => E) 
  ('Post::'a => 'a => heap => E) (h0:heap) (h1:heap) => 
    (HeapInv h0 h1 
     && ((Bad h0 h1 && (forall (r0:'a) (r1:'a) (h0':heap) (h1':heap).
                          ((Bad h0' h1' && HeapInv h0' h1') ==> 'Post r0 r1 h0' h1')))
         || ('Tx ((fun (r0:'a) (r1:'a) (h0':heap) (h1':heap) => 
                      ((HeapInv h0' h1' && DeltaHeap h0 h1 h0' h1') ==> 'Post r0 r1 h0' h1'))) h0 h1)))
type iRDST :: _ = fun ('a::*) ('Tx::('a => 'a => heap => heap => E) => heap => heap => E) =>
    RDST 'a (WithInv 'a 'Tx)

type iRetTx :: _ = (fun ('a::*) 
                      (x:'a) 
                      ('Post::'a => 'a => heap => heap => E) => 'Post (L x) (R x))
    
type iBindTx :: _ = (fun ('a::*) ('b::*) 
                       ('Tx1::('a => 'a => heap => heap => E) => heap => heap => E)
                       ('Tx2::'a => ('b => 'b => heap => heap => E) => heap => heap => E)
                       ('Post::'b => 'b => heap => heap => E)
                       (h0:heap) (h1:heap) => 
    (HeapInv h0 h1 
     && 'Tx1 (fun (y0:'a) (y1:'a) (h0':heap) (h1':heap) => 
         (HeapInv h0' h1' 
          && DeltaHeap h0 h1 h0' h1'
          && (forall (x:'a). ((L x)=y0 && (R x)=y1) ==> 
                'Tx2 x ((fun (z0:'b) (z1:'b) (h0'':heap) (h1'':heap) => 
                            (HeapInv h0'' h1''
                             && DeltaHeap h0' h1' h0'' h1''
                             && 'Post z0 z1 h0'' h1''))) h0' h1')))))


(* val returnRst : x:'a -> Rst 'a (retTx 'a x) *)
(* (\* Commenting out implementations for now, would need to add it to exec *\) *)
(* (\* let returnRst x = fun () -> x *\) *)

(* type bindTx :: _ = fun ('a::*\) ('b::*\)  *)
(*                        ('Tx1::('a => 'a => heap => heap => E) => heap => heap => E) *)
(*                        ('Tx2::'a => ('b => 'b => heap => heap => E) => heap => heap => E) *)
(*                        ('Post::'b => 'b => heap => heap => E) =>  *)
(*     ('Tx1 (fun (y0:'a) (y1:'a) (lh:heap) (rh:heap) => forall (x:'a). ((L x)=y0 && (R x)=y1) ==> 'Tx2 x 'Post lh rh)) *)

(* val bindRst : 'a::* -> 'b::* *)
(*         -> 'Tx1::('a => 'a => heap => heap => E) => heap => heap => E *)
(*         -> 'Tx2::('a => ('b => 'b => heap => heap => E) => heap => heap => E) *)
(*         -> c1:Rst 'a 'Tx1 *)
(*         -> c2:(x:'a -> Rst 'b ('Tx2 x)) *)
(*         -> Rst 'b (bindTx 'a 'b 'Tx1 'Tx2) *)
(* (\* *)
(* let bindRst c1 c2 = *)
(*  (\* 'Tx1 (\y0 y1. forall x. (L x)=y0 && (R x)=y1 ==> 'Tx2 x 'Post) *\) *)
(*   let x = c1 () in *)
(*  (\* 'Tx2 x 'Post *\) *)
(*     c2 x () *)
(*  (\* 'Post *\) *)
*)

val alloc : x:'a -> Rst (ref 'a) (fun ('Post::ref 'a => ref 'a => heap => heap => E) (h0:heap) (h1:heap) => 
    forall (r0:ref 'a) (r1:ref 'a). (not(InDomain r0 h0) && not(InDomain r1 h1))
                                ==> 'Post r0 r1 (Update h0 r0 (L x)) (Update h1 r1 (R x)))

val read : x:ref 'a -> Rst 'a (fun ('Post::'a => 'a => heap => heap => E) (h0:heap) (h1:heap) => 
    forall (y0:'a) (y1:'a). ((y0=Select h0 (L x)) && (y1=Select h1 (R x)))
                                ==> 'Post y0 y1 h0 h1)

val write : x:ref 'a -> v:'a -> Rst unit (fun ('Post::unit => unit => heap => heap => E) (h0:heap) (h1:heap) => 
    'Post () () (Update h0 (L x) (L v)) (Update h1 (R x) (R v)))
