module RO

open Bytes 

let eta = 16 (* fixed size for keys and tags *)

type key = b:bytes { Length(b) = eta } 
type tag = b:bytes { Length(b) = eta }

(* an abstract relational predicate on correctly-generated entries *) 

type safe :: key => key => tag => tag => E 

type alloc = 
  | Hon: alloc 
  | Adv: alloc (* a "ghost" field recording the first allocator for a given key *) 

type state =
  { bad: bool; (* set iff any allocation has failed, e.g. bumped into the other's table *) 
    log: map key (alloc * tag) } (* the map ensures at most one entry per key *)

(* an inductive invariant; unsure about the syntax *) 

type Ok log0 log1 = 
  | Null: Ok [] [] 
  | ConsH: k0:key -> k1:key-> t0:tag -> t1:tag { Safe k0 k1 t0 t1 } -> log0:state -> log1:prestate { Ok log0 log1 } -> Ok ((k0,(Hon,t0))::log0) ((k1,(Hon,t1))::log1)
  | ConsA: k:key ->           t:tag                                 -> log0:state -> log1:prestate { Ok log0 log1 } -> Ok ((k ,(Adv,t ))::log0) ((k, (Adv,t ))::log1)


type goodstate = s:state {| (((L s).bad) = true) || (((R s).bad) = true) || (Ok ((L s).log) ((R s).log)) |} 

val s: ref goodstate
let s = ref []

val sample_hon: k:key -> t:tag {| safe (L k) (R k) (L t) (R t) |} 
val sample_adv: unit -> eq tag 

val hash_hon:  
  k:key -> 
  t:tag {| (L s).bad = true || (R s).bad = true || safe (L k) (R k) (L t) (R t) |}

val hash_adv:  
  k:key {| (L s).bad = true || (R s).bad = true || (L k) = (R k) |} -> 
  t:tag {| (L s).bad = true || (R s).bad = true || (L t) = (R t) |}

let hash_hon k =
  match assoc k !log with 
  | Some (Hon,t) -> t 
  | Some (Adv,t) -> s := { bad = true;   log = !s.log }
  | None         -> let t = sample_hon k in
                    s := { bad = !s.bad; log = (k,(Hon,t)::!s.log) }
                    t

(* there is a variant with additional data for hon, 
   such that we treat Some(Hon,t) also as a collision *)

let hash_adv k =
  match assoc k !log with 
  | Some (Adv,t) -> t 
  | Some (Hon,t) -> s := { bad = true;   log = !s.log }
  | None         -> let t = sample_adv () in
                    s := { bad = !s.bad; log = (k,(adv,t)::!s.log) }
                    t

(* we'll need to bound the probability of collisions; 
   in hash_hon, assuming k is fresh random, this is #adv/#keys 
   in hash_adv, assuming the honest domain is secret random, this is #hon/#keys *)


(* PRIOR ATTEMPTS 

type prestate =
  { LH : (list key); (* honest domain    *)
    LA : (list key); (* adversary domain *)
    log: (list (key * tag)); }

(* we can bound the probability of bad from length LH and length LA *)

type state = s:prestate 
  {| (* Bad *) ( exists k. 
                        ( (k in (L s).LA && k in (L s).LH)) || 
                        ( (k in (R s).LA && k in (R s).LH)) )
  || (* Invariant up to Bad *) 
   (   ( forall k0. k0 in (L s).LH <==> 
           exists t0 k1 t1.  k1 in (R s).LH && (k0,t0) in (L s).log && (k1,t1) in (R s).log && safe k0 k1 t0 t1)
   && 
       ( forall k1. k1 in (R s).LH <==> 
           exists t0 k0 t1.  k0 in (L s).LH && (k0,t0) in (L s).log && (k1,t1) in (R s).log  && safe k0 k1 t0 t1)
   && 
       ( forall k. k in (L s).LA <==> exists t . (k,t) in (L s).log && (k,t) in (R s).log 
   &&
       ( forall k. k in (R s).LA <==> exists t . (k,t) in (L s).log && (k,t) in (R s).log 
))) |} 

let log = ref [] 
type bad = ... 
// must have the same concrete code 
val sample_hon: k:bytes -> t:tag {| safe k t |} 
val sample_adv: unit -> eq tag 

val hash:     k:key -> t:tag {| bad h0 h1 || safe (L k) (R k) (L t) (R t)    |}
val hash_adv  k:key -> t:tag {| bad h0 h1 || (L k) = (R k) ==> (L t) = (R t) |}

let hash k =
  LH := k::!LH;
  match assoc k !log with 
  | Some t -> t
  | None   -> let t = sample_hon b in
              log := (k,t)::!log;
              t

let hash_adv k =
  LA := k::!LA; 
  match assoc k !log with 
  | Some t -> t
  | None   -> let t = sample_adv b in
              log := (k,t)::!log;
              t

(** take 2 **)

type alloc = LH | LA (* first allocator *) 
type prestate =
  { bad: bool; (* has any allocation failed, e.g. bumped into the other's table? *) 
    log: map key (alloc * tag) } (* the map ensures at most one entry per key *)

type state = s:prestate 
  {| ((L s).bad = true || (R s).bad = true) 
  || (* Invariant up to bad *) 
   (   ( forall k0 t0. exists k1 t1. 
         ( (k0,(LH,t0)) in (L s).log ==> (k1,(LH,t1)) in (R s).log && safe k0 k1 t0 t1))
   &&  ( forall k1 t1. exists k0 t0. 
         ( (k1,(LH,t1)) in (R s).log ==> (k0,(LH,t0)) in (L s).log && safe k0 k1 t0 t1))
   &&   ( forall k t. ( (k,(LA,t)) in (L s).log <==> (k,(LA,t)) in (R s).log )))
  |}

(** take 3 **)

type alloc = LH | LA (* first allocator *) 
type prestate =
  { bad: bool; (* has any allocation failed, e.g. bumped into the other's table? *) 
    log: map key (alloc * tag) } (* the map ensures at most one entry per key *)

type state = s:prestate 
  {| ((L s).bad = true || (R s).bad = true) 
  || (* Invariant up to bad *) 
   (   ( forall k0 t0. ( (k0,(LH,t0)) in (L s).log ) ==> 
         exists k1 t1. ( (k1,(LH,t1)) in (R s).log && safe k0 k1 t0 t1 ))
   &&  ( forall k1 t1. ( (k1,(LH,t1)) in (R s).log ) ==> 
         exists k0 t0. ( (k0,(LH,t0)) in (L s).log && safe k0 k1 t0 t1 ))
   &&  ( forall k t.   ( (k ,(LA,t )) in (L s).log <==> (k,(LA,t)) in (R s).log )))
  |}

(* we could avoid the forall/exists by having the two maps have equal structure up to bad; 
   note that the two lists do *not* necessarily have the same structure after bad *)

type state = s:prestate { s.bad = true || Ok ((L s).log) ((R s).log) } 

type Ok log0 log1 = 
  | Null: Ok [] [] 
  | ConsH: k0:key -> k1:key-> t0:tag -> t1:tag { Safe(k0,k1,t0,t1) } -> log0 -> log1 { Ok(log0,log1) }-> Ok ((k0,t0)::log0) ((k1,t1)::log1)
  | ConsA: k:key -> t:tag -> log0 -> log1 { Ok log0 log1 } -> Ok ((k,t)::log0) ((k,t)::log1)

*)
