module Nonce
open Bytes
open CCA2
type nonce = b:bytes { Length b = 16 } (* we need fixed lengths for encryption *)
type principal = eq prin
type tag = eq (b:bytes { Length b = 8 }) 

logic val Hello: tag
logic val Ack  : tag
logic val Error: tag
assume neq Hello Ack

(*$ we need them; unclear why. *)
assume (Length Hello) = 8
assume (Length Ack  ) = 8
assume (Length Error) = 8

(* begin events for messages *)
type SayHello:: id => id => nonce => E
type SayAck::   id => id => nonce => nonce => E 
(* type Pat:: id => nonce => nonce => nonce => nonce => E *)
type Pat:: id => bytes => bytes => bytes => bytes => E

(* specifying authentication properties associated with encryption payload formats *) 
type SentTo :: _ =
  fun (x:id) (p:plain) =>
    (  (forall (a:id) (na:nonce).             
          ((p = Concat Hello (Concat a na)) ==> SayHello a x na))
        
       && (| forall (b:id) (lna:nonce) (rna:nonce) (lnb:nonce) (rnb:nonce).
           (Eq (L p) (Concat Ack (Concat lna lnb)) 
            && Eq (R p) (Concat Ack (Concat rna rnb))
            && (lna <> rna)
            && Safe b
            && SayHello (L x) b lna 
            && SayHello (R x) b rna
            && Pat b lna rna lnb rnb) (* just as a q-inst hint *)
         ==> (SayAck (L x) b lna lnb && SayAck (R x) b rna rnb)|))

let cca : cca2 SentTo = mkCCA2<SentTo> ()
val parse_hello: p:plain -> option(a:prin * na:nonce   { p = Concat Hello (Concat a na) })
val parse_ack:   p:plain -> option(na:nonce * nb:nonce { p = Concat Ack   (Concat na nb) })
val format: t:tag -> b0:bytes -> b1:nonce -> p:plain   { p = Concat t (Concat b0 b1) } 

val nonce_hello: 
  a:prin {Safe a} 
  -> b:prin 
  -> na:nonce { SayHello a b na && 
                (| (Safe(L a) && Safe(L b)) <==> neq nonce (L na) (R na) |)}  

(* provable from the absence of collisions on na *)
assume SayHello_injective:(forall a b0 b1 na. (SayHello a b0 na && SayHello a b1 na) ==> b0 = b1)


val encrypt_hello:
  a:prin { Safe a }
  -> b:prin { Safe b }
  -> cipher
let encrypt_hello a b =
  let na = nonce_hello a b in
  let p = format Hello a na in
  cca.encrypt b p

val decrypt_hello:
  b:prin {Safe b}
  -> c:cipher 
  -> option(a:prin * na:nonce { SayHello a b na || | L a = R a && L na = R na |})
let decrypt_hello b c = 
  match cca.decrypt b c with
    | Some p -> 
      (match parse_hello p with
        | Some((a,na)) -> Some (a,na) 
        | None -> None)
    | None -> None 

val nonce_ack: 
  a:prin 
  -> b:prin {Safe b} 
  -> na:nonce 
  -> nb:nonce { SayAck a b na nb  && 
                (| (Safe(L a) && Safe(L b)) <==> L nb <> R nb |)}  

val encrypt_ack: 
  a:prin 
  -> b:prin {Safe b}
  -> na:nonce { (Safe a && SayHello a b na && (| neq (L na) (R na)|)) 
                || (| Eq (L na) (R na)|)} 
  -> cipher
let encrypt_ack a b na = 
  let nb = nonce_ack a b na in
  let p = format Ack na nb in
  (* assert (SentTo a p); *)
  cca.encrypt a p

val decrypt_ack: 
  a:prin {Safe a}
  -> b:prin
  -> na:nonce   
  -> c:cipher { SayHello a b na && (| Safe (L b) <==> neq nonce (L na) (R na) |)} 
  -> option(nb:nonce{Safe b ==> SayAck a b na nb})  
let decrypt_ack a b na c = 
  match cca.decrypt a c with
    | Some p -> 
      (match parse_ack p with 
        | Some na_nb -> 
            let (na',nb) = na_nb in 
              if na = na' 
              then 
                (assume (| Pat (L b) (L na') (R na') (L nb) (R nb)|); (* pattern inst *)
                 Some nb)
              else None
        | None -> None)
    | None -> None


  
