#monadic(RDST, iRetTx, bindTx)
(*
  A tiny demo for monadic relational inference. 

  The #monadic directive above specifies:
     Rst the monad to use (see relmon.fst)
     retTx, the specification of the monadic return
     bindTx, the specification of the monadic bind

  Aside from the single-sided rules that are specific to the
  relational monad, inference is parametric with regard to the choice
  of the monad.

  Limitations: 

  -- I don't implement a single-sided rule yet. So, currently the
  relational inference for programs with branches is unsound. But, see
  lang-sec/eflow/relational-iflow.txt for a nice single-sided rule
  based on self-composition. Will implement this soon.

*)

module InfoFlowST

(* 
   ll_and_hh is the type of a pure function 
   that maps low inputs to low outputs 
         and on high inputs does whatever
*)

type pure_ll_and_hh = x:int -> RDST int (fun ('Post::int => int => heap => heap => E) h0 h1 => 
    (forall y0 y1. ((L x) = (R x) ==> y0 = y1) ==> 'Post y0 y1 h0 h1))

(* You can write down a let binding and leave its type out. 
   We'll infer a precise type for it. 
   If you write down a val decl with a type, we'll check that the inferred
   type is a subtype of the annotated type. 

   For now, all except the last annotation is commented, but you can
   uncomment them if you like.
*)

(* val id : ll_and_hh *)
(* inferred: *)
(* val id : x:int -> RDST int (fun ('Post::int => int => E) => 'Post (L x) (R x)) *)
let id x = x

val incr : pure_ll_and_hh
(* inferred *)
(* val incr: x:int -> RDST int (fun ('Post::int => int => E) => *)
(*     'Post ((L x) + 1) ((R x) + 1)) *)
let incr x = x + 1

(* Here's the type of a pure function that always returns a low result *)
type pure_any_to_low = x:int -> RDST int (fun ('Post::int => int => heap => heap => E) h0 h1 =>
    (forall y. 'Post y y h0 h1))

(* val zero : any_to_low *)
let zerof x = x - x

(* The main interesting part of the demo is this: 
   We don't need to annotate the types of zero, incr, id etc. 
   But, we can write the function foo below, 
   infer a type for it, and check that it always returns low *)

val foo : pure_any_to_low
let foo x = (zerof (incr (id x))) - 1  

(* val bar : any_to_low  *) (* should fail if you use this type *)
let bar x = (zerof (incr (id x))) + x
  
val sample: x:int -> RDST int (fun ('Post::int => int => heap => heap => E) h0 h1 => 
  (forall y0 y1. ((Eq ((L x) + y0) ((R x) + y1)) ==> 'Post y0 y1 h0 h1)))

val noise_to_declassify: pure_any_to_low
let noise_to_declassify x = (sample x) + x

(* val incr_ref : x:ref int -> RDST unit (fun ('Post::unit => unit => heap => heap => E) h0 h1 =>  *)
(*     (forall h0' h1'. ((Select h0 (L x) = Select h1 (R x)) ==> (Select h0' (L x)) = (Select h1' (R x))) *)
(*        ==> 'Post () () h0' h1')) *)

(* val incr_ref : x:ref int -> RDST unit (fun ('Post::unit => unit => heap => heap => E) h0 h1 => *)
(*     'Post () () (Update h0 (L x) (Add (Select h0 (L x)) 1)) *)
(*                 (Update h1 (R x) (Add (Select h1 (R x)) 1))) *)
let incr_ref x = write x ((read x) + 1)

(* val decr_ref : x:ref int -> RDST unit (fun ('Post::unit => unit => heap => heap => E) h0 h1 => *)
(*     'Post () () (Update h0 (L x) (Sub (Select h0 (L x)) 1)) *)
(*                 (Update h1 (R x) (Sub (Select h1 (R x)) 1))) *)
let decr_ref x = write x ((read x) - 1)

(* Z3 encoding for reasoning about integers in heap is busted *)
val id_ref : x:ref int -> RDST unit (fun ('Post::unit => unit => heap => heap => E) h0 h1 =>
    ('Post () () 
       (UpdHeap h0 (L x) (SelHeap h0 (L x)))
       (UpdHeap h1 (R x) (SelHeap h1 (R x)))))
let id_ref x = 
  let _ = incr_ref x in
    decr_ref x
