module Group 
open Bytes

type elt (* elements of the group *)
type bigint = int
type intmod (n:bigint) = i:bigint{0 <= i && i < n}

logic val idelt:elt 
logic val q:bigint (* q is the order *)
logic val g:elt    (* generator *)
logic val h:elt    (* generator *)

logic val Mlt: elt -> elt -> elt
logic val Exp: elt -> nat -> elt
(* TODO: Axioms *)

logic val alpha:intmod q{h = Exp g alpha}

type text = intmod q (* exponents for values *)
type opng = intmod q (* exponents for openings *)

type Opening :: _ = fun (lx:text) (rx:text) (lo:opng) (ro:opng) => 
    (Eq (lx + (alpha * lo)) (rx + (alpha * ro))) (* operations in modulo q *)

type opening (x:text) = o:opng {| Opening (L x) (R x) (L o) (R o) |}
val opening : x:text -> opening x (* Instance of sample ... TODO *)

logic val Commit: text -> opng -> elt
define Commit_def: forall (x:text) (o:opng). Commit x o = Mlt (Exp g x) (Exp h o)
                                 
val commit: x:text
  -> (o:opng *
        c:eq elt { (| Opening (L x) (R x) (L o) (R o) |) && Commit x o = c })

(*$ TODO, add impl
let commit pp x r = pp.g ^ x * pp.h ^ r 
*)

(* val verify: *)
(*      x:text *)
(*   -> r:opng *)
(*   -> c:elt *)
(*   -> b:bool{ b=true ==> c = Commit pp x r } *)
(* (\*  *)
(* let verify = c = pp.g ^ x * pp.h ^ r  *)
(* *\) *)

(* (\* this is true only computationally;  *)
(*    we should instead specify some computational assumption using up to bad.  *)

(*    Given p g h, the computational challenge is to provide  *)
(*    x0 o0 \neq x1 o1 such that g^x0 * h^o0 = g^x1 * h^o1 *)
(*  *\) *)
(* assume Commit_injective:  *)
(*   forall pp x0 x1 r0 r1.  *)
(*   (Commit pp x0 r0 = Commit pp x1 r1) ==> x0=x1 *)

(* Can prove this post-condition after implementing commit as in the comment above *)
val mult: c0:elt -> c1:elt -> c:elt {forall pp x0 x1 r0 r1.
    (c0=Commit pp x0 r0 && c1=Commit pp x1 r1) ==>
     c=Commit pp (Add x0 x1) (Add r0 r1)}

val exp: c0:elt -> z:int -> c:elt {forall pp x0 r0.
    c0=(Commit pp x0 r0) ==>
    c=Commit pp (Mlt x0 z) (Mlt r0 z)}

end 

(** Metering code **)

module MLT
open Group
open Bytes

type Good :: intmod q => intmod q => E
logic val Hash : bytes -> bytes -> bytes
logic val HT0 : bytes
logic val EltBytes : text -> bytes
(* Axioms .. *)

(* Signing. 

   We could move this to another generic module, 
   taking Signed as a parameter,
   but then we would have to explicitly turn pp and cs into bytes *)
type Signed :: _ = fun (h:bytes) => 
    (exists (x:text) (y:text) (ox:opng) (oy:opng). 
       h=Hash HT0 (Concat (EltBytes (Commit x ox)) (EltBytes (Commit y oy)))
       && Good x y)
       
type dsig = eq bytes 
val sign: h:bytes{Signed h} -> dsig
val verify: 
     h:bytes 
  -> dsig
  -> b:eq bool{ b=true ==> Signed h}

(* SETUP *)
type committed (t:text) = (x:text{x=t} * 
                             o:opng *
                             c:eq elt { (| Opening (L x) (R x) (L o) (R o) |) && Commit x o = c })
val setup: x:text -> y:text{Good x y} -> (committed x * commited y * dsig)
let setup x y =
  let ox = opening () in
  let Cx = Mlt (Exp g x) (Exp h ox) in
  let oy = opening () in
  let Cy = Mlt (Exp g y) (Exp h oy) in
  let tag = sig_sign sk (Hash HT0 (Concat (EltBytes Cx) (EltBytes G Cy))) in
  ((x, ox, Cx), (y, oy, Cy), tag)
          
val meter: pp:pparam
  -> xs:list int{ Readings xs }
  -> (xrs:openings {xs = fsts xrs} *
        cs:eq (list elt) *
        d:dsig{ Commits pp xrs = cs })
let meter pp xs =
  let xrs,cs = commits pp xs in
  (xrs, cs, sign pp cs)

(* USER *)
val sums: xrs:openings
  -> ps:eq (list int)(*{Length ps = Length xrs}*) (*$ todo ... add me *)
  -> (x:int { x = ScalarProduct (fsts xrs) ps } * opening x)
let rec sums xrs ps =
  match xrs,ps with
  | (xr::xrs), (p::ps) -> let x,r = xr in
                          let x0,r0 = sums xrs ps in
                          (x0 + x*p), (r0 + r*p)
  | []       , []      -> 1, 0
(*| _        , _       -> failwith "wrong lengths" *) (*$ todo ... add me *)


(* We actually do not need Readings and Rates to make the payment,
   although they matter to authorize the declassification.  *)
val make_payment: pp:pparam
  -> xrs:openings{ Readings (fsts xrs) }
  -> ps:eq (list int){ Rates ps  
                       &&  (| ScalarProduct (fsts (L xrs)) (L ps) = ScalarProduct (fsts (R xrs)) (R ps) |) }
  -> (eq int * eq int)
let make_payment pp xrs ps =
  let x,o = sums xrs ps in
    (x,o)
      
(* VERIFIER *)
val verify_payment: pp:pparam
  -> ps:eq (list int){Rates ps}
  -> cs:eq (list elt)
  -> s:dsig
  -> x:text
  -> r:opng
  -> b:bool{ b=true ==> exists xs. Readings xs && x=(ScalarProduct xs ps)}
let verify_payment pp ps cs s x r =
  if verify_meter_signature pp cs s then
    (* assert (exists xrs. Readings (fsts xrs) && Eq (list elt) cs (Commits pp xrs)); *)
    let c = scalarExp<Readings> pp cs ps in
      verify pp x r c
  else
    false
      
