// (c) Microsoft Corporation 2009.
module Microsoft.FStar.Fs7

(* Known Ambiguities in Grammar:
    1. type unit = Unit  parses as a type abbreviation.
       qual_def sets this right by looking up Unit
    2. in messages Unit parses as a variable, not as a constructor 
       qual_message sets this right by looking up Unit    
    3. in messages x.y parses as a qualified variable, not as a record lookup
       qual_message sets this right by looking up Unit
*)  
exception Impos

val string_of_int: int -> string
val int_of_string: string -> int
val debug: bool ref
val nsdebug: bool ref
val ask: bool ref
val printPub: bool ref

type pos = string       (* Range information: has the form "filename(line,col-line,col)" *)
type typevar = string   (* Type variable: has the form "'abcde" *)
type lvar = string      (* Local variable used in lets, types, formulas: has the form "abcde" *)
type var = (string list)(* Qualified variable used in constructors, top level functions and values *)
                        (* Has the form: namespace. ... .module.valname *)

val emptyPos:pos
type pub = Private | Public | Default  (* Used to generate fsi's and to check that the attacker interface is public *)
type usage = Function | Predicate | Constructor | Value
                                       (* Used to separate functions from predicates *)
type Qual = {pub : pub; usage : usage}
    
type Constant =                        (* Might consider adding unit, booleans, and bytestrings *)
    String of string
  | Int of int                     
  | Bool of bool
  | Float of float

type Message =                        
    MVar of var * pos                  (* Variable x *)
  | Wild of pos                        (* Anonymous variable _ *)
//  | Name of lvar                       (* Name n: Do we need this? *)
  | Literal of Constant                (* Constant c *)
  | Tuple of Message list              (* Tuple (m1,m2,...,mn) *)
  | Term of var * Message              (* Constructor application h(m1,m2,...,mn) *)
  | Recd of (var * Message) list       (* Records {l1=m1;l2=m2...;ln=mn} *)
  | Lookup of Message * var            (* Record lookup m.l *)
and Pattern = Message 

type Formula =                           
    True                               (* True *)
  | False                              (* False *)
  | Not of Formula                     (* not f *)
  | Iff of Formula * Formula           (* f1 <=> f2 *)
  | Imp of Formula * Formula           (* f1 => f2 *)
  | Conj of Formula * Formula          (* f1 /\ f2 *)
  | Disj of Formula * Formula          (* f1 \/ f2 *)
  | Says of Message * Formula          (* M says f: Unused for now *)
  | All of lvar * Type * Formula * Pattern list       (* !x:T. f *)
  | Ex of  lvar * Type * Formula * Pattern list      (* ?x:T. f *)
  | Pred of var * Message list         (* h(m1,...,mn) *)
  | Eq of Message * Message            (* m1 = m2 *)
  | NEq of Message * Message           (* m1 <> m2 *)
and Type =                                            (* Types T *)
    TVar of typevar                                   (* 'a *)
  | TApp of Type list * Message list * var            (* (T1,...,Tn;M1,...,Mm) F *)
  | TTuple of (lvar * Type) list * Formula            (* <x1:T1,...,xm:Tm>{C} *)
  | TFun of lvar * Type * Formula * lvar * Type * Formula (* [x1:T1]{C1}x2:T2{C2} *)
  | TTop                                              (* Top *)
  | TUvar of Uvar_basis Unionfind.uvar
and Uvar_basis = 
  | FreeVars of typevar list * var list * bool
  | Fixed of Type


type constType = CFun | CPred | CUnknown    (* Used for SMP, CF's new code makes this unnecessary *)

type Sigma = (var * Type * constType) list  (* h1 of T1 | ... | hn of Tn *)

type Kind = Plus          (* Covariant *)
            | Minus       (* Contravariant *)
            | Zero        (* Invariant *)

type TypeDef =                                        (* type definitions: type TypeDef *) 
    TAbs of (typevar * Kind) list * (lvar * Type) list *  var  (* abstract type: ('a1,...,'an;x1:T1,...,xm:Tm) F *)
  | TAlg of (typevar * Kind) list * (lvar * Type) list * var * (var * Type * constType) list
                                                      (* algebraic type: ('a1,...,'an;x1:T1,...,xm:Tm) F = h1 of T1 ...*)
  | TAbbrv of (typevar * Kind) list * (lvar * Type) list * var * Type
                                                      (* type abbrev: ('a1,...,'an;x1:T1,...,xm:Tm) F = T *)
  | TRecd of (typevar * Kind) list * (lvar * Type) list * var * (var * Type) list
                                                      (* records: ('a1,...,'an;x1:T1,...,xm:Tm) F = {l1:T1;...;ln:Tn} *)

type Decl = 
    Val of Qual * var * (typevar * Kind) list * (lvar * Type) list * Type
                                                         (* variable binding: v:('a1...'an;x1...xn) T (ignore Kind) *)
  | Form of pub * var * Formula                          (* policy formula *)
  | Query of pub * var * Formula                                     (* top level query *)
  | AbsType of (typevar * Kind) list * (lvar * Type) list *  var  (* abstract type: see TAbs above *)
  | ConcType of Qual * (typevar * Kind) list * (lvar * Type) list * var * Sigma  
                                                         (* concrete type: see TAlg above *)
  | TypeAbbrv of (typevar * Kind) list * (lvar * Type) list * var * Type
                                                         (* type abbreviation: see TAbbrv above *)
  | TypeRecd of (typevar * Kind) list * (lvar * Type) list * var * (var * Type) list (* Records see TRecd above *)

  | RecType of Qual * TypeDef list                       (* mutually recursive type definitions
                                                            this generalizes the tree typedefs above.
                                                            KB is moving the codebase to use only RecTypes *)
                                                            
  | Exn of var * Type                                    (* exception v of T *)
  | Open of var                                          (* open module *)
  | Kind of Type * Kind                                  (* Type -> Kind *)
  | SubTy of Type * Type                                 (* Type <: Type *)

type fs7 = FS7 of var * string * Decl list                         (* module M d1 d2 ... dn *)

val new_lvar: unit -> lvar
val new_nvar: lvar -> lvar
val new_tvar: unit -> typevar
val new_labelvar: var -> lvar
val lvarToString: lvar -> string
val varToString: var -> string
val messageToString: Message -> string
val kindToString: Kind -> string
val typeToString: Type -> string
val formulaToString: Formula -> string
val moduleToString: var -> string
val compress: Type -> Type
val typevarToString: typevar -> string
val declToString: Decl -> string
val declListToString: (Decl list -> string)
val fvars_message: Message -> var list
val fvars_formula: Formula -> var list
val fvars_type: Type -> var list


(* val qualify_decls: Typecheck.module_env -> Decl list  -> Typecheck.module_env*(Decl list) *)
(* val qualify_fs7: Typecheck.module_env -> fs7 -> Typecheck.module_env*(Decl list) *)
(* val type_to_ty: Type -> Fsharp.typ *)
(* val ty_to_type: Fsharp.typ -> Type *)



val tnames_type: Type -> var list
val tvars_type : Type -> typevar list
val dtvars_type : Type -> typevar list
val all_tvars_type : Type -> typevar list
val tvars_env : Decl list -> typevar list
val uvars_type : Type -> Type list
val fresh_uvar : Type -> Type

(* factor these into a util.fsi *)
val listminus : 'a list -> 'a list -> 'a list when 'a:equality
val union : 'a list -> 'a list -> 'a list when 'a:equality
val intersect : 'a list -> 'a list -> 'a list when 'a:equality

val toPubFs7: fs7 -> string

val error: string -> 'a
val errorPos: string -> string -> 'a
val warn: string -> unit
val warnPos: string -> string -> unit
val conj:Formula * Formula -> Formula
val disj:Formula * Formula -> Formula
val imp:Formula * Formula -> Formula
val notf:Formula -> Formula
val ex:lvar * Type * Formula -> Formula
val all:lvar * Type * Formula -> Formula
