﻿// See LICENSE-js2ml.txt for terms and conditions
// (c) Joel Weinberger 2012. All rights reserved.

module JSMinSyntax

open Common
open System.Collections.Generic

type builtin = JavaScriptSyntax.constant
type prefixOp = JavaScriptSyntax.prefixOp
type infixOp = JavaScriptSyntax.infixOp

type constant =
    | CThis
    | CArguments
    | CBuiltIn of builtin

type value =
    | VConstant of pos * constant
    | VId of pos * id
    (* VLambda is a position, list of tuples of argument pretty print identifiers and the unique, underlying identifier *)
    | VLambda of pos * id list * stmt (* position * args * body *)
    | VObject of pos * (string * exp) list
    | VGlobal of pos
    | VLocals of pos * string (* the string value is the unique identifying scope suffix *)
(*
 * These aren't really all "BuiltInOps", strictly speaking. These are really built in, JavaScript system
 * operations that are internally called.
 *)
and BuiltInOp =
    | BToInt32 of exp
    | BToNumber of exp
    | BToObject of exp
    | BToString of exp

    | BDelete of exp * exp
    | BTypeOf of exp
    | BValueOf of exp

    | BAbstractEquality of exp * exp
    | BRefEquality of exp * exp
    | BIn of exp * exp
    | BInstanceOf of exp * exp
    | BAddition of exp * exp

and exp =
    | EValue of pos * value
    | EOp1 of pos * prefixOp * exp
    | EOp2 of pos * infixOp * exp * exp
    | EDot of pos * exp * id (* position * object expression * property identifier *)
    | EBracket of pos * exp * exp (* position * object expression * property expresionn *)
    | ECall of pos * exp * exp option * exp list (* The first exp is the function expression, the second is the receiver object, and the third is the list of arguments *)
    | ECallDom of pos * exp * exp * exp list * id (* The same as ECall except that the function called *may* be a DOM function *)
    | ECallInvariant of pos * id * id (* A call to the special __invariant__ function that takes an invariant identifier and a locals identifier and applies this to the following while loop. *)
    | EBuiltIn of pos * BuiltInOp
    (* Why are these assignments expression forms? *)
    | EAssign of pos * id * exp
    | EDotAssign of pos * exp * id * exp (* position * object expression * field id * assigning expression *)
    | EBracketAssign of pos * exp * exp * exp (* position * object expression * field expression * assigning expression *)
    | ENew of pos * exp * exp list
    | EObjLiteral of pos * (id * exp) list
    | EList of pos * exp * exp (* Evaluates a series of expressions, returning the last one in the list as the value of the entire expression. *)
        (* This one too ... why not a statement? *)
    | EIf of pos * exp * exp * exp
    | EVar of pos * id * exp * exp (* position * let binding id * expression to bind to * expression to evaluate with the new binding *)

and stmt =
    | SSeq of pos * stmt * stmt
    | SIf of pos * exp * stmt * stmt
    | SWhile of pos * exp * stmt
    | SReturn of pos * exp
    | SBreak of pos
    | SEmpty of pos
    | SExp of exp
    | SForInStmt of pos * id * exp * stmt (* position * variable to assign to * object expression * body *)
    | SVar of pos * id * exp * stmt (* position * let binding id * expression to bind to * stmt to evalutate with the new binding *)
    | SVarDecl of pos * id  (* position * let binding id * expression to bind to * stmt to evalutate with the new binding *)
    | SThrow of pos * exp
    | SComment of string
    | SCommentSmall of string

type prog =
    | PStmt of pos * stmt

val js_to_jsmin_raw : JavaScriptSyntax.prog -> prog

val JSSyntax2JSMinMap : Dictionary<System.Object, System.Object>
